# encoding=utf-8

from base64 import b64decode as real_b64decode, b64encode as real_b64encode
import binascii
from collections import namedtuple
from collections.abc import Mapping, Sequence
from datetime import datetime, timezone
from io import RawIOBase
import math
import re
import struct

from ijson import basic_parse as ijson_basic_parse, items as ijson_items, parse as ijson_parse
import iso8601
from logbook import Logger


log = Logger(__name__)


# signal value used when parsing an invalid date
INVALID_DATE = datetime.min.replace(tzinfo=timezone.utc)


# typing is a bit bugged; b64encode uses bytes for both input and output, convert output to str
def b64encode(s):
    """
    Encodes `bytes` to a `str` using default Base64 encoding.

    :param s: `bytes`
    :return: *s*, Base64-encoded
    :rtype: `str`
    """
    return str(real_b64encode(s), 'ascii')


def b64decode(s, validate=False):
    """
    Decode a Base64-encoded string.

    :param s: the string to decode
    :param validate: validate the input against the Base64 alphabet
    :return: decoded byte string
    :rtype: `bytes`
    :raise TypeError: when an invalid character is found
    """
    if not isinstance(s, bytes):
        s = bytes(s, 'ascii')

    try:
        return real_b64decode(s, validate=validate)
    except binascii.Error as e:
        raise TypeError(str(e)) from e


def omit_empty(mapping):
    """
    Filters mapping to a new dict with non-``None`` values.

    :param mapping: dict to filter
    :return: dict containing keys and values from mapping that are non-empty
    """
    return {key: value for key, value in mapping.items() if value is not None}


def glue_url(*path):
    """
    Glues parts of a url path to each other using /s, dropping any steps that
    are None.

    :param path: steps to join
    :return: a url path
    """
    return '/'.join(str(step).strip('/ ') for step in path if step is not None)  # omit steps that are None


def json_events(fd):
    """
    Generates prefixed json parsing events. See also ijson.parse.

    :param fd: file-like object to read from
    :return: an event generator
    """
    return ijson_parse(ijson_basic_parse(fd, use_float=True))


def json_items(events, prefix):
    """
    Generates python values parsed from the provided event stream that
    conform to prefix. See also ijson.items.

    :param events: a json event stream (see json_events)
    :param prefix: prefix of elements to generate (dot-separated, use "item"
                   for array items)
    :return: a value generator
    """
    return ijson_items(events, prefix)


# binary suffixes for byte sizes
_byte_sizes = ('bytes', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB')


def format_byte_size(size, template='{value:.4g} {unit}'):
    """
    Converts a byte size into a human-readable format using binary suffixes.

    :param size: the value to be formatted
    :param template: a format string with two named parameters: *value* and
        *unit*
    :return: a human-readable file size
    """
    # find the order of magnitude (log_2(size) / 10)
    order = int(math.log(size, 2) // 10) if size else 0

    if order >= len(_byte_sizes):
        # exceeding ludicrous file sizes, default to bytes
        return '{} bytes'.format(size)
    # format the template with a human-readable byte size
    return template.format(value=size / (1 << (order * 10)), unit=_byte_sizes[order])


def parse_byte_size(value):
    """
    Turns a human-readable byte size into a numeric value.

    Suffixes are always interpreted as binary and can take a single or triple
    character form (e.g. *m*, *MiB* for megabytes), casing is ignored.

    :param value: a human-readable file size
    :return: an `int`-representation of *value*
    """
    try:
        # match '15', '15k', '15KiB', '1.2 M', …
        match = re.match(r'^(?P<num>(?:\d*\.)?\d+)\s*(?P<suffix>[a-z]*)$', value, re.IGNORECASE)
        num, suffix = match.groups()
        num = float(num)
        # default order 0 (just bytes)
        order = 0
        if suffix:
            # there was a suffix in value, determine the order by looking at the suffix' index in _byte_sizes
            order = next((i for i, size in enumerate(_byte_sizes) if size.lower().startswith(suffix.lower())))

        # calculate the byte size as the product of the supplied value and the binary order multiplier
        return round(num * (1 << (order * 10)))
    except Exception as e:
        # any exception (when match of next fails, for example) result in an invalid byte size error
        raise ValueError('invalid byte size: {}'.format(value)) from e


def format_datetime(ts):
    """
    Coverts a `datetime.datetime` object into ISO-8601 wire format, while
    requiring a timezone.

    :param ts: the `datetime` to be converted
    :return: *ts*, in ISO-8601 format
    """
    if not ts.tzinfo:
        raise ValueError('datetime without timezone')

    return ts.isoformat()


def parse_datetime(value):
    """
    Parses an ISO8601-formatted timestamp into a `datetime` object.

    Returns `.INVALID_DATE` when an invalid value is passed (e.g. a negative
    year component, which is not supported by `datetime`).

    :param value: an ISO8601-formatted `str`
    :return: a `datetime` object representing the same value as *value*
    """
    try:
        return iso8601.parse_date(value)
    except iso8601.ParseError:
        # NB: value "-1234-01-21…" would technically be valid in ISO 8601, though Python's datetime does not allow it
        log.warning('failed to parse {} as datetime, defaulting to {}', value, INVALID_DATE, exc_info=True)
        return INVALID_DATE


class GeographicLocation(namedtuple('GeographicLocation', ('latitude', 'longitude'))):
    """
    Describes a 2-tuple *(latitude, longitude)* to translate location data to
    and from Hansken's wire format.
    """

    _pattern = re.compile(r'^(?P<latitude>[+\-]?\d+\.\d+)(?P<longitude>[+\-]?\d+.\d+)/?$')

    @classmethod
    def from_string(cls, value):
        """
        Parses a string value into a `GeographicLocation`.

        :param value: a string-representation to be parsed
        :return: a value of type `cls`, parsed from *value*
        """
        match = cls._pattern.match(value)
        if not match:
            raise TypeError('invalid {cls.__name__}: {value}'.format(cls=cls, value=value))

        # use keyword arguments to instantiate cls (extensions beware :))
        return cls(latitude=float(match.group('latitude')), longitude=float(match.group('longitude')))

    def __str__(self):
        # use zero-padded floating point values with forced sign rounded to a 5 digit fraction
        return '{location.latitude:+09.5f}{location.longitude:+010.5f}'.format(location=self)


class Vector(Sequence):
    """
    Describes a sequence of floating point numbers to translate vector data to
    and from Hansken's wire format.
    """

    # compile a struct for values in a Vector: big-endian 32 bit float
    _struct = struct.Struct('>f')

    @classmethod
    def from_base64(cls, value):
        """
        Creates a `.Vector` from *value*, assuming it will be base64-encoded
        sequence of big-endian 4-byte IEEE-754 floating points.

        :param value: a base64-representation to be parsed
        :return: a new `.Vector`
        """
        return cls(b64decode(value))

    @classmethod
    def from_sequence(cls, seq):
        """
        Creates a `.Vector` from *seq*, assuming it will be a sequence of
        elements compatible with the ``f`` struct format character.

        :param seq: a sequence of `float`
        :return: a new `.Vector`
        """
        size = cls._struct.size
        # allocate a buffer that would fit the entire sequence
        buffer = bytearray(len(seq) * size)
        for idx, element in enumerate(seq):
            # pack all elements into the pre-allocated buffer
            cls._struct.pack_into(buffer, idx * size, element)

        # wrap buffer into bytes to make sure the result's _raw attribute is a bytes (required by __bytes__)
        return cls(bytes(buffer))

    def __init__(self, data):
        """
        Creates a new `.Vector` backed by the provided data.

        :param data: a byte buffer, aligned to 4 bytes
        """
        if len(data) % self._struct.size:
            raise TypeError(f'misaligned data, {len(data)} not a multiple of {self._struct.size}')

        self._raw = data

    def __len__(self):
        return len(self._raw) // self._struct.size

    def __getitem__(self, item):
        size = self._struct.size

        if isinstance(item, slice):
            # or self[idx], but bypass a bunch of type and size checks
            return [self._struct.unpack_from(self._raw, offset=idx * size)[0]
                    # let range and slice generate the actual indices we need
                    for idx in range(*item.indices(len(self)))]

        if not isinstance(item, int):
            # mimic the error list would create
            raise TypeError(f'indices should integers or slices, not {item.__class__.__name__}')

        if item >= len(self):
            # we're responsible for this check (__iter__ relies on this behaviour)
            raise IndexError

        # unpack a single value from the underlying buffer
        # (unpack always returns a tuple, we always need the first and only element)
        return self._struct.unpack_from(self._raw, offset=item * size)[0]

    def __eq__(self, other):
        # define equality as the same type and identical data
        return isinstance(other, Vector) and self._raw == other._raw

    def __hash__(self):
        return hash((13, self._raw))  # avoid a hash collision with the hash of _raw

    def __bytes__(self):
        return self._raw

    def __str__(self):
        return b64encode(bytes(self))

    def __repr__(self):
        return f'{self.__class__.__name__}({list(self)})'


class MultiContext:
    """
    Context manager delegating its `__enter__` and `__exit__` methods to the
    provided set of context managers.

    `__enter__` is called on all delegates in the order they've been supplied.
    `Exception` s raised from these calls are collected, if any of the
    delegates raised an `Exception`, a `ValueError` is raised with an error
    message and the recorded `Exception` instances. `__exit__` is *not* called
    in this case.

    `__exit__` is called on all delegates in the the reverse order they've been
    supplied. The exception detail arguments for `__exit__` are passed to the
    delegates verbatim. The return value for `__exit__` will be truthy if any
    of the delegate calls return a truthy value (which would cause the
    interpreter to suppress the error.

    Any delegates raising an `Exception` of their own on `__exit__` produce
    undefined behaviour; errors on `__exit__` should be avoided.

    Use `.MultiContext` as a context manager:

    .. code-block:: python

        with MultiContext(manager1, manager2):
            do_things()

        # note that this is equivalent to
        with manager1, manager2:
            do_things()

        # MultiContext can be used as a return type for multiple context managers
        with create_context_managers():
            do_things()
    """

    def __init__(self, *contexts):
        """
        Creates a new `.MultiContext`, delegating its `__enter__` and
        `__exit__` calls to the provided delegate context managers.

        :param contexts: delegate context managers
        """
        self.contexts = contexts

    def __enter__(self):
        errors = []

        for context in self.contexts:
            try:
                context.__enter__()
            except Exception as e:
                # warn here to indicate which context failed to enter
                log.warn('context {} failed to __enter__', context, e)
                errors.append(e)

        if errors:
            # don't silence errors
            raise ValueError('failed to __enter__ MultiContext', *errors)

        # always return self to make sure self.__exit__ is called
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # default to falsy return value
        suppress = False

        for context in reversed(self.contexts):
            # allow short circuiting the or, but make sure delegate.__exit__ is called
            suppress = context.__exit__(exc_type, exc_val, exc_tb) or suppress

        return suppress


class Namespace:
    """
    Utility class to create a tiny namespace with some simple attributes.
    """

    def __init__(self, **kwargs):
        for attr, value in kwargs.items():
            setattr(self, attr, value)

    def extend(self, **kwargs):
        for attr, value in kwargs.items():
            setattr(self, attr, value)


class DictView(Mapping):
    """
    Dict-like class with separator-aware value retrieval. A dot (.) is used as
    the default separator. Values can be retrieved in one of three methods
    (the first two of which are like a normal dict):

    - calling get: view.get('some.property')
    - indexing: view['some.property']
    - attribute access: view.some.property

    If the retrieved property is a dict, it is wrapped in a DictView to enable
    further DictView access methods.
    """

    # sentinel value used to check when a KeyError should be raised rather than returning a value
    _no_default = object()

    def __init__(self, source, separator='.'):
        self._source = source
        self._separator = separator

    def get(self, key, default=None):
        try:
            value = self._source
            for step in key.split(self._separator):
                value = value[step]

            if isinstance(value, dict):
                return DictView(value)
            else:
                return value
        except KeyError:
            return default

    def __len__(self):
        return len(self._source)

    def __iter__(self):
        return iter(self._source)

    def __contains__(self, item):
        return self.get(item) is not None

    def __getitem__(self, item):
        value = self.get(item, self._no_default)
        if value is self._no_default:
            raise KeyError(item)
        else:
            return value

    def __str__(self):
        return str(self._source)


def view_with_attrs(name=None, attrs=()):  # default attrs to an empty tuple (no attributes)
    """
    Creates a class definition that is an extension of DictView. Attribute
    names are cached and used to lazily get them (either as their original
    name or the PEP8-style equivalent) from the instance's source when
    accessed using __getattr__. For use in interactive shells, __dir__
    includes the named attributes translated to PEP8-style names.

    :param name: name of the generated view class
    :param attrs: attributes to be defined for the requested class
    :return: extension of DictView that loads attributes from the provided
        source dict, renamed to python's conventions
    """

    # transform a sequence of attribute names to a mapping of {name: factory}
    # factory None will not touch the value
    if not isinstance(attrs, Mapping):
        attrs = {attr: None for attr in attrs}

    def to_value(value, name):
        if value is None:
            # avoid handing factories None, None is identical to 'not set'
            return value

        factory = attrs.get(name)
        # let factory transform the value if available
        return factory(value) if factory else value

    class AttrView(DictView):
        def __init__(self, source):
            super().__init__(source)

        def __getattr__(self, item):
            if item in attrs:
                # attribute is defined, fetch from parent
                return to_value(self.get(item), item)

            prop = to_property_name(item)
            if prop in attrs:
                # translated attribute is defined, fetch the translation from parent
                return to_value(self.get(prop), prop)

            # no such attribute
            raise AttributeError(item)

        def __dir__(self):
            base = set(super().__dir__())
            # extend the set of attrs reported up the mro with those explicitly defined for this AttrView
            # NB: the result of dir()/__dir__ is not guaranteed to be 'correct', it is meant for interactive inspection
            return sorted(base | {to_attr_name(attr) for attr in attrs})

    if name is not None:
        AttrView.__name__ = name

    return AttrView


def to_property_name(name):
    """
    Converts a name to a property name by convention (*camelCase*) by
    converting *snake_case* to *snakeCase*.

    :param name: name to convert
    :return: *name* converted to a property name
    """
    return re.sub(r'_[a-z]', lambda match: match.group()[1].upper(), name)


def to_attr_name(name):
    """
    Converts a name to an attribute name by convention (*snake_case*) by
    converting *camelCase* to *camel_case*.

    :param name: name to convert
    :return: *name* converted to a attribute name
    """
    return re.sub(r'[a-z][A-Z]', lambda match: '_'.join(match.group().lower()), name).lower()


def to_class_name(name):
    """
    Converts a name to a class name by convention (*PascalCase*) by converting
    *snake_case* to *SnakeCase*.

    :param name: name to convert
    :return: *name* converted to a class name
    """
    name = re.sub(r'_[A-z]', lambda match: match.group()[1].upper(), name)
    return name[0].upper() + name[1:]


class ChunkedIO(RawIOBase):
    """
    Turns a generator of chunks into raw I/O, which can in turn be fed to
    something like an `io.BufferedReader`.

    .. code-block:: python

        # requesting an imaginary big file, to be sent chunked
        response = requests.get('http://big-files.net/big-file.bin',
                                stream=True)  # make sure to request a stream
        # using 8KiB bytes chunk size, turn the response into a BufferedReader
        reader = io.BufferedReader(
            ChunkedResponseIO(response.iter_content(8192))
        )
    """

    def __init__(self, chunks):
        """
        Creates a `.ChunkedIO`, reading chunks of `bytes` from *chunks*.

        .. note::

            `.ChunkedIO` does no type checking on *chunks* or the chunks it
            yields; only `bytes` chunks will likely work as expected.

        :param chunks: an iterable yielding `bytes`
        """
        self._chunks = chunks
        self._current = None

    def seekable(self):
        return False

    def readable(self):
        return True

    def writable(self):
        return False

    def readinto(self, b):
        if self.closed:
            return 0

        max_read = len(b)
        num_read = 0

        try:
            # reuse current (partial) chunk or use the next chunk if the current one is None or empty
            self._current = self._current or next(self._chunks)
            # keep reading till num_read reaches max_read
            while num_read < max_read:
                # split current chunk into a buffer to be used right now and one for the next call
                buf, self._current = self._current[:max_read - num_read], self._current[max_read - num_read:]
                # put the data to be used now into b
                b[num_read:num_read + len(buf)] = buf
                num_read += len(buf)
                # reuse current (partial) chunk or use the next chunk if the current one is None or empty
                self._current = self._current or next(self._chunks)
        except StopIteration:
            # self._chunks is exhausted, mark self._current as invalid
            self._current = None

        return num_read

    def close(self):
        self._chunks.close()
        super().close()


def flatten_mapping(mapping, separator='.', prefix=None):
    """
    Flattens *mapping* into a single level `dict` by concatenating nested keys.

    :param mapping: the mapping to be flattened
    :param separator: separator to be used to concatenate nested keys
    :param prefix: prefix for all keys in the nested result (typically only
        useful for recursive calls)
    :return: *mapping*, flattened into a single level `dict`
    """
    # collect dict items (tuples) in a list
    items = []

    for key, value in mapping.items():
        if prefix:
            # update the key to be used if there's a prefix
            key = separator.join((prefix, key))

        try:
            # attempt recursive flattening (…)
            items.extend(flatten_mapping(value, prefix=key).items())
        except AttributeError:
            # (…) not a mapping, apparently, just track this item
            items.append((key, value))

    # return a flattened version of mapping
    return dict(items)



# encoding=utf-8

from http.server import HTTPServer, SimpleHTTPRequestHandler
import re
from threading import Thread
import time

from logbook import Logger
import requests


from hansken.util import ChunkedIO

log = Logger(__name__)


UPLOAD_PATH = '/image-upload'
PATH_PATTERN = re.compile(r'^/(?:image-upload/)?(?P<image_id>[a-f0-9\-]+)(?P<extension>\.nfi(:?\.idx)?)(:?\?.*)?$',
                          re.IGNORECASE)
PATH_TEMPLATE = '/image-upload/{image_id}{extension}'


def read_chunked(fobj):
    # read hexadecimal number of bytes in the next chunk
    chunk_size = int(fobj.readline(), 16)
    while chunk_size:
        yield fobj.read(chunk_size)
        # consume \r\n at the end of chunk
        fobj.readline()
        # read hexadecimal number of bytes in the next chunk
        chunk_size = int(fobj.readline(), 16)


class WebHDFSRequestHandler(SimpleHTTPRequestHandler):
    """
    HTTP request handler that mimics WebHDFS's handling of PUT requests.
    """
    def log_message(self, format, *args):
        # avoid super writing to stderr
        log.debug('{} - {}', self.address_string(), format % args)

    def do_redirect(self):
        path = PATH_PATTERN.match(self.path)
        if path:
            location = PATH_TEMPLATE.format(**path.groupdict())
            log.debug('sending temporary redirect from {} to {}', self.path, location)
            self.send_response(307)
            self.send_header('Location', location)
            self.end_headers()
            return None
        else:
            log.debug('sending bad request (cannot match path {} for redirect)', self.path)
            self.send_response(400)
            self.end_headers()
            return None

    def _slice_chunk(self, response, image_id, chunk, offset, extension, at_attempt):
        # NB: method uses plain raise; should only be called from an except block!
        if response.status_code == 416:
            # TODO: return a 416 response for applicable upload issues rather than a 500 (HANSKEN-17174)
            resource_size = response.headers['Content-Range']
            resource_size = re.match(r'^bytes .+/(?P<size>\d+)$', resource_size, re.IGNORECASE)
            if not resource_size:
                log.warn('no usable resource size available in 416 response: {}', response.headers['Content-Range'])
                raise
            # reported size does not match offset, skip the difference
            skip = int(resource_size.group('size')) - offset
        elif response.status_code == 500:
            response_text = response.text.strip()
            # using the plain text error message to determine how much of the bytes were processed previously
            # despite the error response
            offsets = re.match(r'^expected offset (\d+) but was (\d+)$', response_text, re.IGNORECASE)
            if offsets:
                expected, sent = offsets.groups()
                # skip the difference, send the trailing end of the buffer we tried to append
                skip = int(expected) - int(sent)
            else:
                log.warn('failed to recognize offsets for partial retry from error message "{}"', response_text)
                # raise original error with original traceback (causing a 500, see note below)
                raise
        else:
            # cannot deal with this error in a satisfiable way, continue original error
            raise

        if skip < 0:
            # not something we can solve by skipping bytes, continue original error
            raise

        # offset + skip should equal expected, return (recursive) result (not resetting the attempt
        # counter) or bubble an error
        log.warn('retrying partial chunk at offset {}, skipping {} bytes that seem to be present at offset {}',
                 offset + skip, skip, offset)
        return self.process_chunk(image_id, chunk[skip:], offset + skip, extension, at_attempt=at_attempt)

    def process_chunk(self, image_id, chunk, offset, extension, at_attempt=None):
        def should_retry():
            return self.server.max_retries is None or attempt <= self.server.max_retries + 1

        def should_slice(response):
            return (
                # trigger 1: response 416 (range not satisfiable)
                response.status_code == 416 or (
                    # trigger 2: response 500 (internal error) with a specific error message body
                    response.status_code == 500 and response.text.strip().lower().startswith('expected offset')
                )
            )

        attempt = at_attempt or 1
        while attempt == 1 or should_retry():
            try:
                return self.server.upload_callback(image_id=image_id, extension=extension,
                                                   data=chunk, offset=offset)
            except requests.ConnectionError as e:
                attempt += 1
                if should_retry():
                    log.warn('uploading a chunk at offset {} failed due to {}, retrying in {}s',
                             offset,
                             str(e),
                             self.server.retry_wait,
                             e)
                    time.sleep(self.server.retry_wait)
            except requests.HTTPError as e:
                attempt += 1
                response_text = e.response.text.strip()
                if should_slice(e.response):
                    return self._slice_chunk(e.response, image_id, chunk, offset, extension, at_attempt=attempt)
                if e.response.status_code == 500 and should_retry():
                    log.warn('uploading a chunk at offset {} failed due to {}: {}: {}, retrying in {}s',
                             offset,
                             e.response.status_code, e.response.reason, response_text,
                             self.server.retry_wait,
                             e)
                    time.sleep(self.server.retry_wait)
                else:
                    # chunk upload is not being retried, log exception
                    log.exception('uploading a chunk at offset {} failed: {}: {}: {}',
                                  offset,
                                  # log detailed error information, including error response body
                                  e.response.status_code, e.response.reason, response_text,
                                  e)
                    # raise original error with original traceback (causing a 500, see note below)
                    raise

        # loop exits only at exhausted number of attempts, give up
        # NB: this error will show up on the 'client side', i.e. the hansken-image-tool process as a 500 response
        raise ValueError('maximum attempts ({}) reached for chunk at offset {} for image {}'.format(
            attempt, offset, image_id
        ))

    def do_upload(self):
        path = PATH_PATTERN.match(self.path)
        if path:
            image_id = path.group('image_id')
            extension = path.group('extension')
            data = self.rfile
            if self.headers.get('Transfer-Encoding') == 'chunked':
                log.debug('getting chunked data, wrapping rfile with chunk generator')
                data = read_chunked(self.rfile) if self.server.streaming else ChunkedIO(read_chunked(self.rfile))

            if self.server.streaming:
                log.info('image will be uploaded as a stream')
                try:
                    log.debug('forwarding image data as-is to remote')
                    self.server.upload_callback(image_id=image_id, extension=extension, data=data)
                except (requests.ConnectionError, requests.HTTPError) as e:
                    if hasattr(e, 'response'):
                        log.exception('uploading image data failed, no retry available: {}: {}: {}',
                                      e.response.status_code, e.response.reason, e.response.text.strip(),
                                      e)
                    else:
                        log.exception('uploading image data failed, no retry available: {}', str(e), e)

                    # close i/o resources (force broken pipe on the client side)
                    data.close()
                    self.rfile.close()
                    # remote service can't process our request, indicate gateway error to client
                    log.debug('sending bad gateway response for image {}', image_id)
                    self.send_response(502)
                    self.end_headers()
                    return None
            else:
                log.info('image will be uploaded in chunks')
                buffer = memoryview(bytearray(self.server.bufsize))
                num_read = data.readinto(buffer)
                offset = 0

                while num_read:
                    try:
                        log.debug('uploading {} bytes of {} image data at offset {}', num_read, extension, offset)
                        self.process_chunk(image_id, buffer[:num_read], offset, extension)
                        offset += num_read
                        num_read = data.readinto(buffer)
                    except ValueError:
                        # close i/o resources (force broken pipe on the client side)
                        data.close()
                        self.rfile.close()
                        raise

            log.debug('sending created response for image {}', image_id)
            self.send_response(201)
            self.end_headers()
            return None
        else:
            log.debug('sending bad request (cannot match path {} for upload)', self.path)
            self.send_response(400)
            self.end_headers()
            return None

    def do_PUT(self):  # noqa: N802
        log.debug('got PUT request for {}', self.path)

        try:
            if self.path.startswith(UPLOAD_PATH):
                return self.do_upload()
            else:
                return self.do_redirect()
        except Exception as e:
            log.exception('handling PUT request for {} failed', self.path, e)
            self.send_response(500)
            self.end_headers()
            return None


class WebHDFSServer(HTTPServer):
    """
    HTTP server that mimics WebHDFS on a local address.
    """
    def __init__(self, upload_callback, streaming, bufsize=8 << 20, max_retries=8, retry_wait=10.0):
        # bind to port 0, let OS find a free port
        super().__init__(('localhost', 0), WebHDFSRequestHandler)

        self.upload_callback = upload_callback
        self.bufsize = bufsize
        self.max_retries = max_retries
        self.retry_wait = retry_wait
        self.streaming = streaming
        self._thread = None

    def __enter__(self):
        if not self._thread:
            self._thread = Thread(name='webhdfs-server-{}'.format(self.server_port),
                                  target=self.serve_forever)
            log.info('starting WebHDFS server on port {} in the background', self.server_port)
            self._thread.start()

            return super().__enter__()
        else:
            raise ValueError('server already started or in error state')

    def __exit__(self, exc_type, exc_val, exc_tb):
        log.info('shutting down WebHDFS server on port {}', self.server_port)
        self.shutdown()
        # wait for server thread to exit
        self._thread.join()
        self._thread = None

        return super().__exit__(exc_type, exc_val, exc_tb)



# encoding=utf-8

from collections import namedtuple
from collections.abc import Mapping
from enum import Enum, unique
from itertools import chain
import json
import warnings

from logbook import Logger

from hansken import fetch
from hansken.abstract_trace import AbstractTrace
from hansken.query import Term, to_query
from hansken.util import (b64decode, b64encode,
                          DictView, flatten_mapping,
                          format_datetime, GeographicLocation, INVALID_DATE,
                          parse_datetime, to_attr_name, Vector, view_with_attrs)


log = Logger(__name__)


_converter = namedtuple('_converter', ('serialize', 'deserialize'))


# collection of wire / python type converters by model type
CONVERTERS = {
    # convert binary to / from base64
    'binary': _converter(
        b64encode,
        b64decode,
    ),
    # convert dates to / from datetime.datetime objects (*requiring* a timezone)
    'date': _converter(
        format_datetime,
        parse_datetime,
    ),
    # convert latLong to / from GeographicLocation
    'latLong': _converter(
        str,  # trust we'll hit GeographicLocation.__str__ or coerce a str to itself
        GeographicLocation.from_string,
    ),
    # convert vector to / from Vector
    'vector': _converter(
        str,  # trust we'll hit Vector.__str__ or coerce a str to itself
        Vector.from_base64,
    ),
}


def image_from(trace_uid):
    warnings.warn('image_from is deprecated, use image_from_uid', DeprecationWarning)
    return image_from_uid(trace_uid)


def image_from_uid(trace_uid):
    """
    Splits *trace_uid* into its two parts, *image* and *id*, returning the
    first.
    Note that a `.Trace` object will provide these as properties `image_id` and
    `id`.

    :param str trace_uid: an image uid
    :return: the image UUID from *trace_uid*
    :rtype: `str`
    """
    return trace_uid.split(':', 1)[0]


def image_from_trace(trace):
    """
    Attempts to get an image id from *trace*, whether *trace* is a `.Trace`
    object or `dict`-like.

    :param trace: a `.Trace` or `dict`-like trace
    :return: the image UUID from *trace*
    :rtype: `str`
    """
    if hasattr(trace, 'image_id'):
        return trace.image_id
    if 'image' in trace:
        return trace.get('image')
    if 'uid' in trace:
        return image_from_uid(trace.get('uid'))

    raise ValueError('unable to get image id from trace {}'.format(trace))


@unique
class Privileged(Enum):
    """
    Possible privileged states of a `.Trace`. Values that correspond to 'not
    privileged' (`None` or `.rejected`) are falsy, making them suitable to
    check whether a trace is privileged.
    """

    suspected = 'suspected'  #: trace is suspected of being privileged
    confirmed = 'confirmed'  #: trace is confirmed to be privileged
    rejected = 'rejected'  #: trace is confirmed to be *not* privileged

    def __bool__(self):
        # rejected value should be falsy (see class doc)
        return self is not Privileged.rejected

    def __eq__(self, other):
        if isinstance(other, str):
            # enable comparisons with str values
            return str(self) == other

        return super().__eq__(other)

    def __str__(self):
        return self.value


class Snippet(DictView):
    """
    Snippet result, enabling rendering of a highlighted snippet of text
    content. Usable as a dictionary where key ``'content'`` contains a snippet
    of text and key ``'highlights'`` contains a list of dictionaries encoding
    highlighted terms in the content.
    """

    def render(self, start='[[', end=']]'):
        """
        Render this snippet by surrounding highlights with *start* and *end*
        marker strings, e.g.:

        .. code-block:: python

            >>> my_snippet.render()
            'A [[snippet]] with the term "[[snippet]]" highlighted.'
            >>> my_snippet.render(start='<em>', end='</em>')
            'A <em>snippet</em> with the term "<em>snippet</em>" highlighted.'

        :param start: start marker around highlights
        :param end: end marker around highlights
        :return: this `.Snippet`, highlighted as a `str`
        """
        content = self.get('content')
        highlights = self.get('highlights')

        # gather chunks of text to be joined together later
        chunks = []

        # start at text offset 0
        prev_end = hl_end = 0
        for highlight in highlights:
            hl_start, hl_end = highlight['start'], highlight['end']

            # for each highlight, add content from last highlight up to current start (…)
            chunks.append(content[prev_end:hl_start])
            # (…) then a highlight start marker (…)
            chunks.append(start)
            # (…) followed by the actual highlight content (…)
            chunks.append(content[hl_start:hl_end])
            # (…) and a highlight end marker
            chunks.append(end)

            prev_end = hl_end

        # add a final chunk from the end of the last highlight to the end of the content string
        chunks.append(content[hl_end:])

        return ''.join(chunks)

    def __str__(self):
        return self.get('content')


class Trace(AbstractTrace):
    """
    Base class for traces. Defines convenience methods to navigate or
    manipulate a trace. Trace data may be accessed using `open <.Trace.open>`.
    """

    # value separating sequence numbers in trace ids (e.g. 0-1-23)
    ID_SEP = '-'
    # signal value for lazy post-init values
    _UNINITIALIZED = object()

    def __init__(self, source, context=None):
        super().__init__(source)

        # a super will likely also set these attributes, but we need these for operation without any super classes
        # retrieve the values from source as we're only interested in intrinsics, self.get() inspects the trace model
        self.image_id = source.get('image')  # this will get an intrinsic property, leaving the image type as member
        self.id = source.get('id')
        self.uid = source.get('uid')
        self.name = source.get('name')
        # set parent identifiers if applicable (id/uid might not be available or trace is a root)
        self.parent_id = self.id[:self.id.rindex(Trace.ID_SEP)] if self.id and Trace.ID_SEP in self.id else None
        self.parent_uid = self.uid[:self.uid.rindex(Trace.ID_SEP)] if self.uid and Trace.ID_SEP in self.uid else None
        # default to an empty set, 'type' in trace.types should be possible
        # note that a modeled trace class below would override this
        self.types = self.get('types', set())
        # invocations of the update method are counted, see Trace.update method
        self._update_count = 0

        self._tags = None
        self._notes = None
        self._privileged = Trace._UNINITIALIZED  # None is a valid value, use signal value
        self._audits = None

        self._context = context
        self._project_id = context.project_id if context else None

    @property
    def context(self):
        """
        The `ProjectContext <hansken.remote.ProjectContext>` instance that
        created this `.Trace`.
        """
        if self._context:
            return self._context
        else:
            raise ValueError('project context for trace {} not set'.format(self.uid))

    @property
    def image_name(self):
        """
        The name / description of this `.Trace`'s ``image_id``, or ``None``.
        """
        return self.context.image_name(self.image_id) if self.image_id else None

    @property
    def parent(self):
        """
        This `.Trace`' parent `.Trace`, or ``None`` if not applicable.
        """
        return self.context.trace(self.parent_uid) if self.parent_uid else None

    def note(self, note, refresh=None):
        """
        Add a note to this `.Trace`.

        :param str note: the note itself
        :param refresh: if `True`, force a full project refresh, making this
                        note immediately searchable
        """
        self.context.note(self.uid, note, refresh=refresh)

    @property
    def notes(self):
        """
        The notes attached to this `.Trace`. Note that this does not include
        notes added by `note <.Trace.note>`.
        """
        if self._notes is None:
            # TODO: wrap with something that allows to delete a note (HANSKEN-2247)
            self._notes = self.get('user.annotated.#note') or ()

        return self._notes

    def tag(self, tag, refresh=None):
        """
        Tag this trace.

        :param str tag: the tag to set
        :param refresh: if `True`, force a full project refresh, making this
                        note immediately searchable
        """
        self.context.tag(self.uid, tag, refresh=refresh)

    @property
    def tags(self):
        """
        The tags attached to this `.Trace`. Note that this does not include
        tags added by `tag <.Trace.tag>`.
        """
        if self._tags is None:
            # TODO: wrap with something that allows to delete a tag (HANSKEN-2247)
            self._tags = self.get('user.annotated.tags') or ()

        return self._tags

    @property
    def privileged(self):
        """
        The privileged state of this trace, either `None` or one of
        `.Privileged`. Note that `None` is not a valid value when *setting*
        the `.privileged` attribute, an operation that requires authorization.
        """
        if self._privileged is Trace._UNINITIALIZED:
            self._privileged = self.get('user.annotated.privileged', None)

            if self._privileged:
                # value has been set, turn it into a Privileged
                try:
                    self._privileged = Privileged(self._privileged)
                except ValueError as e:
                    # 'reset' the attribute as if freshly constructed
                    self._privileged = Trace._UNINITIALIZED
                    raise ValueError(
                        'unknown privileged state: {}'.format(self.get('user.annotated.privileged'))
                    ) from e

        return self._privileged

    @privileged.setter
    def privileged(self, status):
        status = Privileged(status)
        self.context.mark_privileged(self.uid, status)
        self._privileged = status

    @property
    def creator(self):
        """
        The tool that created this `.Trace`, or ``None`` if unknown. Includes
        the version of that tool, e.g.: ``toolname 1.2.3``.

        .. note::
            This value is formatted by ``hansken.py``, it is not suitable for
            use with queries (like finding other traces created by the same
            tool).
        """
        # attempt to retrieve new-style creator metadata, fall back to old-style metadata
        tool = self.get('system.processed.origin.createdBy') or self.get('system.processed.tool.meta.creator')
        if tool:
            # provide the creator of this trace including its version
            return '{} {}'.format(tool, self.tool_versions.get(tool) or '(unknown version)')

    @property
    def tool_versions(self):
        """
        The tools and versions that are responsible for this `.Trace`'s
        metadata, as a `dict` mapping the names of tools to their respective
        versions. Tool versions typically include the versions of critical
        software libraries used by those tools.
        """
        # attempt to retrieve new-style tool information
        toolruns = self.get('system.processed.toolrun')
        if toolruns:
            # property should be returning {tool name: tool version}, take version from last occurrence
            return {run.get('tool'): run.get('version') for run in toolruns}

        # fall back to old tool metadata (but never None)
        return self.get('system.processed.tool.meta.version') or {}

    @property
    def audits(self):
        """
        An audit log of user-initiated changes to this `.Trace` in the form of
        a sequence of `dict`s, ordered by the audit's creation timestamp. The
        audit log can be empty, but never `None`.
        """
        if self._audits is None:
            # default to empty sequence
            self._audits = self.get('system.processed.audit') or ()
            # createdOn should always be set, but don't crash when it's not available
            self._audits = sorted(self._audits, key=lambda audit: audit.get('createdOn') or INVALID_DATE)

        return self._audits

    def tracelets(self, tracelet_type, query=None, sort=None):
        """
        Provides or retrieves tracelets of type *type*.

        The exact return type of a call to `.tracelets` depends on the tracelet
        type being requested. If the remote defines *type* to be 'few', the
        result will be a `list` of `.Tracelet` objects. If the remote defines
        *type* to be 'many', the result will be a `.SearchResult` of `.Tracelet`
        objects. Note that *query* can only be used with the latter.

        :param tracelet_type: the tracelet type to request
        :param query: query to match tracelets to
        :param sort: ordering of tracelets
        :return: a sized iterable of `.Tracelet` s (iterable once)
        """
        trace_uid_query = Term('traceUid', self.uid)
        # ensure we're searching for tracelets belonging to this trace
        query = to_query(query) & trace_uid_query if query else trace_uid_query
        return self.context.search_tracelets(tracelet_type, query=query, sort=sort)

    @property
    def children(self):
        """
        A `SearchResult <hansken.remote.SearchResult>` instance containing the
        child traces of this `.Trace`, if any.
        """
        return self.context.children(self.uid)

    @property
    def data_types(self):
        """
        A set of data type names available for this `.Trace`. These names can
        be used with calls to `open <.Trace.open>` or attribute access like

        .. code-block:: python

            if 'raw' in trace.data_types:
                # trace has a raw data stream, attribute access to data.raw.size will be safe
                print('raw data size:', trace.data.raw.size)

            for data_type in trace.data_types:
                # format a file name as the trace's name, using the data type name as the extension
                # (e.g. "some-file.raw" or "another-file.text")
                out_file = '{}.{}'.format(trace.name, data_type)
                print('writing first 64 bytes to', out_file)
                with open(out_file, 'wb') as out_file:
                    # out_file now opened for writing in binary mode
                    # write the first 64 bytes of trace's stream of type data_type to the file
                    out_file.write(trace.open(data_type, size=64).read())

        :return: data type names available for this `.Trace` (possibly empty,
            but never `None`)
        :rtype: `set`
        """
        if hasattr(self, 'data'):
            # likely a modeled trace
            return set(self.data.keys())

        # get a view on the extracted data type (possibly empty or None) (…)
        data = self.get('system.extracted.data')
        # (…) and always return a set
        return set(data.keys() if data else [])

    def open(self, stream='raw', offset=0, size=None, key=fetch):
        """
        Open a data stream of a named stream (default ``raw``) for this
        `.Trace`.

        .. note::
            Multiple calls to `read(num_bytes)` on the stream resulting from
            this call works fine in Python 3, but will fail in Python 2.

        :param stream: stream to read
        :param offset: byte offset to start the stream on
        :param size: the number of bytes to make available
        :param key: key for the image of this trace (default is to fetch the
            key automatically, if it's available)
        :return: a file-like object to read bytes from the named stream
        :rtype: `io.BufferedReader`
        """
        return self.context.data(self.uid, stream, offset, size, key)

    @property
    def preview_types(self):
        """
        A set of preview type names (mime types) available for this `.Trace`.
        These names can be used with calls to `preview <.Trace.preview>`.

        :return: preview type names available for this `.Trace` (possibly
            empty, but never `None`)
        :rtype: `set`
        """
        previews = self.get('previews')
        return set(previews.keys() if previews else [])

    def preview(self, mime_type):
        """
        Gets a preview of a particular mime type, e.g. 'text/plain' or
        'image/png'.

        :param mime_type: the preview type to get
        :return: `bytes` or `None`
        """
        previews = self.get('previews')
        if previews and mime_type in previews:
            return b64decode(previews.get(mime_type))
        else:
            # no previews or no such mime type in previews
            return None

    def snippets(self, query, num=100, before=200, after=200):
        """
        Generate snippets surrounding term hits from *query* in any of the
        data streams of this trace.

        :param query: the query to generate snippets for (should contain term
            queries, or no snippets will be generated)
        :param num: maximum number of snippets to return
        :param before: number of bytes to include before the term hits
        :param after: number of bytes to include after the term hits
        :return: `list` of `.Snippet` instances
        """
        with self.context.search(query=Term('uid', self.uid) & to_query(query), count=1, snippets=num) as result:
            snippets = result.takeone(include='snippets')
            if snippets:
                # takeone() is safe, unpacking a potential None is not
                _, snippets = snippets
            if not snippets:
                # either no results or no snippets for the only result
                return []

            # gather unique terms from al of the generated snippets
            terms = list({snippet['term'] for snippet in snippets})
            image_key = self.context.key(self.image_id)
            image_key = b64encode(image_key) if image_key else None
            if hasattr(self, 'data'):
                # likely a modeled trace, let get() deal with data origins and categories
                data_sizes = {data_type: self.get('data.{}.size'.format(data_type))
                              for data_type in self.data_types}
            else:
                # plain trace, assume data property path to be in origin system
                data_sizes = {data_type: self.get('system.extracted.data.{}.size'.format(data_type))
                              for data_type in self.data_types}
            # transform all the snippets into snippet requests, using all the returned terms for each request
            snippets = [
                {'uid': self.uid,
                 'imageKey': image_key,
                 'dataType': snippet['dataType'],
                 'highlights': terms,
                 # request data from *around* the term hit
                 # offset parameters are validated, clip to stream bounds
                 'start': max(0, snippet['start'] - before),
                 'end': min(snippet['end'] + after, data_sizes[snippet['dataType']])}
                for snippet in snippets
            ]

            return self.context.snippets(*snippets)

    def update(self, key_or_updates=None, value=None, data=None, overwrite=False):
        """
        Requests the remote to update or add metadata properties for this
        `.Trace`.

        .. note::

            Calls to `update` will *not* update the source of the `.Trace`
            it's being called on. To get a `.Trace` instance including the
            changes made after a successful call to `update`, use
            ``trace.context.trace(trace.uid)`` to request a new instance of
            a trace with this `.Trace`'s identifier.

            Please note that, for performance reasons, all changes are buffered and not directly effective in
            subsequent search, update and import requests. As a consequence, successive changes to a single
            trace might be ignored. Instead, all changes to an individual trace should be bundled in a single
            update or import request.
            The project index is refreshed automatically (by default every 30 seconds), so changes will become
            visible eventually.

        :param key_or_updates: either a `str` (the metadata property to be
            updated) or a mapping supplying both keys and values to be updated
            (or `None` if only data is supplied)
        :param value: the value to update metadata property *key* to (used
            only when *key_or_updates* is a `str`)
        :param data: a `dict` mapping data type / stream name to bytes to be
            imported
        :param overwrite: whether properties to be imported should be
            overwritten if already present
        :return: processing information from remote
        """
        # count invocations of the update method
        self._update_count += 1

        if self._update_count > 1:
            log.warn("This trace ({}) is being updated more than once. This could result in lost updates, please "
                     "supply a mapping with all batched updates instead of single keys and values.", self.id)

        updates = key_or_updates
        if isinstance(key_or_updates, str):
            updates = {key_or_updates: value}

        return self.context.update_trace(self, updates, data=data, overwrite=overwrite)

    def child_builder(self, name=None):
        """
        Create a `.TraceBuilder` to build a trace to be saved as a child of
        this `.Trace`. Note that ``name`` is a mandatory property for a trace,
        even though it is optional here. A ``name`` can be added later using
        the `.TraceBuilder.update` method. Furthermore, a new trace will only
        be added to the index once explicitly saved (e.g. through
        `.TraceBuilder.build`).

        :param name: the name for the trace being built
        :return: a `.TraceBuilder` set up to create a child trace of this
            `.Trace`
        """
        builder = self.context.child_builder(self.uid)

        if name:
            # optionally add a name for the new child trace
            builder.update('name', name)

        return builder

    def __repr__(self):
        return '<{0.__class__.__module__}.{0.__class__.__name__} {0.uid} ({0.name})>'.format(self)


class IncompleteTracePropertyError(ValueError):
    """
    A `ValueError` raised when a trace property is missing required parts.
    """
    pass


class TraceModel(DictView):
    """
    Utility to deal with intricacies surrounding the trace / data model used
    by Hansken. Used by ``hansken.py`` to translate and validate
    user-specified metadata properties to their corresponding place in the
    data structure for a trace in Hansken.
    """

    def __init__(self, source):
        super().__init__(source)

        self._intrinsics = set(self.get('properties', default=[]))
        self._origins = tuple(sorted(self.get('origins.keys', default={}).keys()))
        # create a mapping tracking the category of a type, for efficiency later
        self._categorized_types = {type_name: category_name
                                   for category_name, category in self.get('origins.categories').items()
                                   for type_name in category.get('types', default={}).keys()}
        self._categorized_properties = {property_name: category_name
                                        for category_name, category in self.get('origins.categories').items()
                                        for property_name in category.get('properties', default={}).keys()}
        # NB: it's possible that there's categories without defined types, _categorized_types.values() would skip those
        self._categories = set(self.get('origins.categories', default={}).keys())
        self._types = set(self._categorized_types.keys())
        # create a mapping tracking the data types for mapped types
        self._mapped_types = {type_name: set(type_details.get('keys'))
                              for category_name, category in self.get('origins.categories').items()
                              for type_name, type_details in category.get('types', default={}).items()
                              # type is mapped if its 'keys' property is non-empty
                              if type_details.get('keys')}
        self._data_types = set(self.get('origins.categories.extracted.types.data.keys', []))

    @property
    def intrinsics(self):
        """
        The intrinsic properties (properties that any trace can have,
        regardless of its type(s)) defined by the trace model.
        """
        return self._intrinsics

    def is_intrinsic(self, steps):
        """
        Checks whether the property defined by *steps* is an intrinsic
        property.

        :param steps: steps through a `.Trace`' data structure
        :return: whether the property defined by *steps* is an intrinsic
            property
        """
        return steps[0] in self.intrinsics

    @property
    def origins(self):
        """
        The origins defined by the trace model, typically *system* and *user*.
        """
        return self._origins

    @property
    def categories(self):
        """
        The categories of types and properties defined by the trace model,
        e.g. *extracted* or *annotated*.
        """
        return self._categories

    @property
    def types(self):
        """
        The trace types defined by the trace model, e.g. *file* or
        *classification*.
        """
        return self._types

    @property
    def data_types(self):
        """
        Data named data types defined by the trace model for the "data" trace
        type.
        """
        return self._data_types

    def _expand_from_type(self, category, type_name, rest):
        # category and type name are already provided, lead with those
        yield category
        yield type_name

        try:
            data_types = self._mapped_types.get(type_name, set())
            if data_types:
                # mapped type, verify the data type is valid for the type (e.g. data.raw)
                data_type = next(rest)
                if data_type in data_types:
                    yield data_type
                else:
                    raise KeyError('unknown data type "{}" for trace type "{}"'.format(data_type, type_name))

            # verify the property name is valid for the type
            type_properties = self.get('origins.categories.{}.types.{}.properties'.format(category, type_name),
                                       default={})
            property_name = next(rest)
            # retrieve property details or trigger a KeyError
            property_details = type_properties[property_name]
            yield property_name

            if property_details.get('isMap', default=False):
                # provide map key for mapped properties (e.g. email.headers.transfer-encoding)
                # glue all remaining steps for the map property together (but require at least one)
                yield self._required_map_key(rest)
        except StopIteration:
            # expected another element from next(rest)
            raise IncompleteTracePropertyError()

    @staticmethod
    def _required_map_key(steps):
        # join the remaining / available steps on a dot
        # this results in 'transfer-encoding' if that's the only remaining elements
        # this results in 'nsrl.os.product' if there's three remaining elements 'nsrl', 'os' and 'product'
        map_key = '.'.join(steps)
        if not map_key:
            # empty string, signal an 'incomplete' trace property
            raise IncompleteTracePropertyError()

        return map_key

    def expand(self, name):
        """
        Expands a trace property to 'steps' through a nested data structure.

        Inserts a properties category if unspecified, does *not* include an
        origin.

        :param name: the property name to expand, excluding an origin
        :return: a `tuple` of 'steps'
        :raise ValueError: when a provided *name* is not defined by the trace
            model or is missing required parts
        """
        try:
            if '.' not in name:
                # either an intrinsic property or categorized property without its category
                if name in self.intrinsics:
                    if self.get('properties.{}.isMap'.format(name), default=False):
                        # no dot, missing map key for mapped property (e.g. previews)
                        raise IncompleteTracePropertyError()

                    # return intrinsic property as a one-tuple (e.g. name)
                    return name,

                # categorized property without the specified category (e.g. tags)
                return self._categorized_properties[name], name

            # split the provided name into steps through the model
            steps = iter(name.split('.'))
            step = next(steps)
            if step in self.intrinsics:
                # mapped intrinsic property (e.g. previews.text/plain)
                return step, self._required_map_key(steps)

            if step in self.categories:
                category = step
                step = next(steps)
                if step in self._categorized_properties:
                    if self._categorized_properties[step] != category:
                        raise ValueError('unknown trace property: "{}"'.format(name))

                    # mapped categorized property (e.g. category.misc.something)
                    if self.get('origins.categories.{}.properties.{}.isMap'.format(category, step), default=False):
                        # glue all remaining steps for the map property together (but require at least one)
                        return category, step, self._required_map_key(steps)
                    # regular categorized property (e.g. annotated.tags)
                    return category, step

                # categorized but not a property, must be a categorized type property
                return tuple(self._expand_from_type(category=category, type_name=step, rest=steps))

            if step in self.types:
                # type without category, find corresponding category and expand from there
                category = self._categorized_types[step]
                return tuple(self._expand_from_type(category=category, type_name=step, rest=steps))
        except (StopIteration, IncompleteTracePropertyError):
            # expected another step, raise new error to include full property name in error message
            raise IncompleteTracePropertyError('incomplete trace property: "{}"'.format(name))
        except KeyError:
            raise ValueError('unknown trace property: "{}"'.format(name))
        else:
            raise ValueError('unknown trace property: "{}"'.format(name))

    def get_serializer(self, steps):
        try:
            if self.is_intrinsic(steps):
                converter = CONVERTERS.get(self.get('properties.{}.type'.format(steps[0])))
                return converter.serialize if converter else None

            model_type = self.get('origins.categories.{}.properties.{}.type'.format(*steps[:2]))
            if model_type:
                converter = CONVERTERS.get(model_type)
                return converter.serialize if converter else None

            if steps[1] in self._mapped_types and steps[2] in self._mapped_types[steps[1]]:
                # steps models a mapped type property, steps[2] is the data type name, omit it to retrieve model type
                model_steps = steps[0], steps[1], steps[3]
            else:
                # steps models a regular type property, model type to be retrieved with the first 3 steps
                model_steps = steps[:3]

            model_type = self.get('origins.categories.{}.types.{}.properties.{}.type'.format(*model_steps))
            if model_type:
                converter = CONVERTERS.get(model_type)
                return converter.serialize if converter else None
        except IndexError:
            raise ValueError('unknown trace property: "{}"'.format('.'.join(steps)))
        else:
            raise ValueError('unknown trace property: "{}"'.format('.'.join(steps)))


class TraceBuilder(DictView):
    """
    Utility class to aid in creating user-defined traces or updating existing
    ones. A `.TraceBuilder` is a trace model aware view on a nested mapping,
    using the trace model to both validate requested updates and finding the
    correct spot for values in the nested mapping.

    This class is not intended for direct user instantiation, see

    - `.Trace.child_builder`
    - `.ProjectContext.child_builder`
    - `.TraceBuilder.child_builder`
    """

    def __init__(self, model, source=None, target=None, context=None,
                 debug=False):
        super().__init__(source or {})

        self._model = model
        self._updates = set()
        self._data = {}
        self._context = context
        # mark debug as a 'public' value, enable user to change it later
        self.debug = debug

        self._uid = None

        if isinstance(target, TraceBuilder):
            # target 'parent' is also a TraceBuilder, defer resolving target arguments to a later time
            self._project_id = self._parent_uid = None
            self._parent_builder = target
        elif target:
            # explicit target, no parent builder
            self._project_id, self._parent_uid = target
            self._parent_builder = None
        else:
            # no known target, no issue yet, maybe user will provide target later
            self._project_id = self._parent_uid = self._parent_builder = None

    def update(self, key_or_updates, value=None):
        """
        Add or overwrite new metadata properties to this builder.

        *key_or_updates* can mix dotted properties and nested structures,
        all keys and values are merged before applying updates. A `.TraceModel`
        is used to find the proper fully qualified property names if needed,
        allowing both e.g. ``update('file.name', 'File Name')`` and
        ``update({'extracted': {'file': {'name': 'file name'}}})``.

        :param key_or_updates: either a `str` (the metadata property to be
            updated) or a mapping supplying both keys and values to be updated
            (or `None` if only data is supplied)
        :param value: the value to update metadata property *key* to (used
            only when *key_or_updates* is a `str`)
        :return: this `.TraceBuilder`
        """
        # merge two ways of calling this method into a single data structure
        updates = key_or_updates
        if isinstance(key_or_updates, str):
            updates = {key_or_updates: value}

        # collect all updates in flattened form
        for key, value in flatten_mapping(updates).items():
            # let the trace model expand the key to be updated (will crash for unknown properties)
            steps = self._model.expand(key)
            path, last_step = steps[:-1], steps[-1]

            # enter the user-specified realm for non-intrinsic properties
            current = self._source
            if not self._model.is_intrinsic(steps):
                current = current.setdefault('user', {})

            # walk through the source to arrive at the last stop
            for step in path:
                current = current.setdefault(step, {})

            # find a serializer and set the actual value where's it's supposed to go
            serializer = self._model.get_serializer(steps)
            current[last_step] = serializer(value) if serializer else value

            # track that we've applied an update for the steps just taken
            self._updates.add('.'.join(steps))

        return self

    def add_data(self, stream, data):
        """
        Add data to this trace as a named stream.

        :param stream: name of the data stream to be added
        :param data: data to be attached
        :return: this `.TraceBuilder`
        """
        if stream not in self._model.data_types:
            log.warn('data type stream name "{}" not defined by model, '
                     'trace being constructed might be invalid', stream)

        self._data[stream] = data

        return self

    @property
    def updates(self):
        """
        A collection of updates tracked by this `.TraceBuilder`.
        """
        # create a deterministic view on the updates this builder has applied to its source
        return tuple(sorted(self._updates))

    @property
    def context(self):
        """
        The `ProjectContext <hansken.remote.ProjectContext>` instance that
        created this `.TraceBuilder`.
        """
        if self._context:
            return self._context
        else:
            raise ValueError('project context for trace builder not set')

    @property
    def target(self):
        """
        The combination of *(project id, parent trace uid)* this
        `.TraceBuilder` applies to.
        """
        if self._project_id and self._parent_uid:
            return self._project_id, self._parent_uid
        elif self._parent_builder is not None:  # NB: an empty parent builder is falsy
            if not self._parent_builder._uid:
                raise ValueError('parent builder has no uid (yet), missing builder target parameters')
            # use parent's project id and parent's uid as the target for this builder
            return self._parent_builder._project_id, self._parent_builder._uid
        else:
            raise ValueError('builder target missing project id and/or parent trace uid')

    def child_builder(self, name=None):
        """
        Creates a new `.TraceBuilder` to build a child trace to the trace to be
        represented by this builder.

        .. note::

            Parent `.TraceBuilder`s should be built using the `.build()` call
            *before* their child builders as the unique trace identifier (uid)
            for the parent is needed to build a child trace.

        :param name: name of the new child trace
        :return: a `.TraceBuilder` set up to save a new trace as the child
            trace of this builder
        """
        # create an instance of our own type, supply all the arguments the child builder needs
        builder = type(self)(model=self._model, target=self, context=self._context, debug=self.debug)

        if name:
            # optionally add a name for the new child trace
            builder.update('name', name)

        return builder

    def build(self):
        """
        Save the trace being built by this builder to remote.

        .. note::

            If this `.TraceBuilder` was put in debug mode, the trace is *not*
            sent to remote but is instead logged at warning level.

        :return: the new trace' uid (or `None` in debug mode)
        """
        if self.debug:
            def value_path(key):
                if self._model.is_intrinsic(self._model.expand(key)):
                    return key
                else:
                    return 'user.{}'.format(key)

            log.warning(
                '{builder.__class__.__module__}.{builder.__class__.__name__} in debug mode, not creating with with '
                'parent {parent}; properties: {updates}; data: {data}',
                builder=self, parent=self.target[1], updates=', '.join(self.updates), data=', '.join(
                    '{}: {} bytes'.format(name, len(stream))
                    for name, stream in sorted(self._data.items())
                ) or '(no data)'
            )
            # collect update values only, apply the inverse of update's handling of intrinsic / user values
            # (updates are collected without the origin prefix)
            update_info = {key: self.get(key) for key in map(value_path, self.updates)}
            # mimic a REST trace, log the trace in JSON format (value serialization is already done by update)
            update_info = json.dumps(update_info, ensure_ascii=False, indent=2)
            log.debug('trace not created: \n{trace}', trace=update_info)

            return None
        else:
            if self._uid:
                # avoid creating duplicates from the same builder
                raise ValueError('builder was already built (resulting uid: {})'.format(self._uid))

            log.debug(
                'submitting {builder.__class__.__module__}.{builder.__class__.__name__} as a new trace with '
                'parent {parent}; properties: {updates}; data: {data}',
                builder=self, parent=self.target[1], updates=', '.join(self.updates), data=', '.join(
                    '{}: {} bytes'.format(name, len(stream))
                    for name, stream in sorted(self._data.items())
                ) or '(no data)'
            )
            self._uid = self.context.connection.create_trace(*self.target, self, data=self._data)
            return self._uid


def value_converter_for(property_details):
    """
    Creates a trace value converter for a provided property value definition.
    Key "type" in *property_details* primarily defines the conversion. The
    following trace value types are converted:

    - *date*: a `datetime.datetime` object with timezone
    - *binary*: a `bytes` object

    If *property_details* states that the the value will be a sequence,
    mapping or mapping to sequences, the created converter will take this into
    account.

    :param property_details: a property definition dict
    :return: a callable that converts a wire format to a python type, or
        ``None`` (no conversion needed or supported)
    """
    def sequence_converter(converter):
        # transform the sequence of raw values to a sequence of transformed values
        return lambda values: [converter(value) for value in values]

    def mapping_converter(converter):
        # transform the mapped values to an equivalent dict of converted values with the same mapping
        return lambda values: {key: converter(value) for key, value in values.items()}

    # converter for value type defined in the trace model
    converter = CONVERTERS.get(property_details.get('type'))
    converter = converter.deserialize if converter else None

    if converter:
        # the value has a converter, check to see if we'll be getting sequences or mappings
        # NB: this means that a mapping of lists is possible ({'key': ['value', 'value', …]})!
        if property_details.get('isList'):
            converter = sequence_converter(converter)
        if property_details.get('isMap'):
            converter = mapping_converter(converter)

    # whether still None or constructed to be something fancy, converter is now what it needs to be
    return converter


def trace_types(model):
    """
    Reduces a trace model dict to the bare essentials used to define the
    properties of a trace, e.g.:

    .. code-block:: python

        {('extracted', 'file'): {'name': None,  # name is a str, no conversion needed
                                 'createdOn': iso8601.parse_date}}

    :param model: the full trace model received from remote
    :return: a mapping that maps a trace type's category and name to its
        properties defined in the model along with a converter for the
        property's type (or ``None``), again as a mapping
    """
    essentials = {}

    for category, types in model['origins']['categories'].items():
        # (some categories can be empty)
        if types:
            # get the name of the trace type and whatever the trace model defines for that type
            for type_name, trace_type in types['types'].items():
                # reduce the details for all the trace type's properties to just its type
                properties = {prop: value_converter_for(details)
                              for prop, details
                              in trace_type['properties'].items()}
                essentials[(category, type_name)] = properties

    return essentials


def expand_types(model, *type_names):
    """
    Creates a `list` of property names for the requested trace types, obtained
    from the provided model. Resulting property names are formatted as
    ``<trace_type>.<property_name>``, e.g. ``'file.createdOn'``.

    :param model: the model to read types and properties from
    :param type_names: the names of the types to be expanded
    :return: a `list` of typed property names
    """
    names = []
    model_types = model['origins']['categories']['extracted']['types']
    for type_name in type_names:
        # generate property names including their type names (sorted to force deterministic behaviour)
        names.extend(sorted('{}.{}'.format(type_name, prop)
                            for prop in model_types[type_name]['properties'].keys()))

    return names


class TraceTypeView(DictView):
    """
    Utility class to turn a trace's type into a prefixed view on the trace's
    source dict.
    """

    def __init__(self, source, category, name, origins=('system', 'user'), converters=None):
        """
        Create a new trace type view.

        :param source: trace source
        :param category: category where the named type resides
        :param name: name of the trace type
        :param origins: origins where the trace type could be sourced from,
            ordered by priority
        :param converters: mapping of property names to optional type
            converter (a `callable`)
        """
        super().__init__(source)

        self._category = category
        self._name = name
        self._origins = origins

        self._converters = converters or {}

    def get(self, key, default=None):
        for origin in self._origins:
            # see if there's a value for <origin>.<category>.<name>.<key>
            value = super().get(self._separator.join((origin, self._category, self._name, key)),
                                default=self._no_default)
            if value is not self._no_default:
                # found a valid value
                converter = self._converters.get(key)
                return converter(value) if converter else value

        # no origin has a value for key in source
        return default

    def keys(self):
        keys = set()
        for origin in self._origins:
            # retrieve trace type for each origin
            origin_type = super().get(self._separator.join((origin, self._category, self._name)))
            if isinstance(origin_type, Mapping):
                # create a union of keys in each origin, if available
                keys.update(origin_type.keys())

        return keys

    def __iter__(self):
        # self._source is not rooted at the trace type, override to use keys at the view point
        return iter(self.keys())

    def __len__(self):
        # self._source is not rooted at the trace type, override to use keys at the view point
        return len(self.keys())

    def __str__(self):
        # NB: comprehensions create their own scope, causing zero-argument super() to fail inside them, create a
        #     super-proxy inside the __str__ def to be able to use it inside the comprehension
        parent = super()
        # super's __str__ would return a str() of the entire source
        # create a str of all the chunks of data that would be available through get(), keyed by the chunk/origin's name
        return str({origin: parent.get('.'.join((origin, self._category, self._name)))
                    for origin in self._origins})

    def __repr__(self):
        return '<{0.__class__.__module__}.{0.__class__.__name__} ({0._name})>'.format(self)


class TraceletTypeView(DictView):
    """
    Utility class to wrap a tracelet value that takes type conversions into
    account.
    """

    def __init__(self, source, category, name, converters=None):
        """
        Create a new tracelet type view.

        :param source: tracelet source
        :param category: category where the named type resides
        :param name: name of the trace type
        :param converters: mapping of property names to optional type
            converter (a `callable`)
        """
        super().__init__(source)

        self._category = category
        self._name = name
        self._converters = converters or {}

    def get(self, key, default=None):
        # no origins or path trickery to apply, _source is the tracelet itself
        value = super().get(key, default=self._no_default)
        if value is not self._no_default:
            converter = self._converters.get(key)
            return converter(value) if converter else value
        else:
            return default


def trace_class_from_model(model):
    """
    Creates a Trace class definition from the provided model. Both intrinsic
    properties and the properties defined for each type are automatically
    defined for the class and will be None should they not be present on the
    actual trace instance.

    The constructor for the resulting class accepts a single argument, being
    a dict as received from remote, with properties defined by model. See
    `.DictView` for access patterns for instances of the resulting class.

    :param model: the full trace model received from remote
    :return: a class definition for a Trace defined by model
    """
    # read intrinsic properties from model
    intrinsics = {prop: value_converter_for(details)
                  for prop, details
                  in model['properties'].items()}
    # create an unnamed view class from the intrinsic properties (used as a super for TraceView, setting intrinsics)
    intrinsic_view = view_with_attrs(attrs=intrinsics)

    # determine the origins defined by the model
    # NB: current client and server implementations don't care about the order defined here, this may change in the
    #     future™, sorted here to ensure deterministic behaviour
    origins = sorted(model['origins']['keys'].keys())
    categories = model['origins']['categories'].keys()

    # create {(category, name) → {prop_name → converter}} mapping from the model
    model_types = trace_types(model)
    # create a reverse lookup to find the category of a type
    types = {name: category for category, name in model_types.keys()}

    # determine plural and mapped trace types
    plural_types = set()
    mapped_types = set()
    for category, type_name in model_types.keys():
        trace_type = model['origins']['categories'][category]['types'][type_name]
        # plural-type defined by few/many cardinality or (legacy) isList: true
        if trace_type.get('cardinality') in ('few', 'many') or trace_type.get('isList'):
            plural_types.add((category, type_name))
        # map-typed defined by "keys" being non-empty
        if trace_type.get('keys'):
            mapped_types.add((category, type_name))

    # create a trace type class for each singular / simple or mapped trace type defined by the model using type:
    # - new type is named alike extracted.text_message (builtin type's module "abc" is prefixed in __repr__)
    # - new type inherits from
    #   - TraceTypeView, able to get() values from a trace's deeply nested source dict
    #   - an AttrView, making attribute access easier, calling get() (defined by TraceTypeView) to get the respective
    #     values
    # - define no additional class variables
    type_classes = {(category, type_name): type('.'.join(map(to_attr_name, (category, type_name))),
                                                (TraceTypeView, view_with_attrs(type_name, properties.keys())),
                                                {})
                    for (category, type_name), properties in model_types.items()
                    if (category, type_name) not in plural_types}
    # analogous to the singular trace types, create Python types for the plural trace types, only difference being the
    # TraceletTypeView over the TraceTypeView, that doesn't insert multi-origin trickery into the get() method (as
    # tracelet values don't support deep paths anyway, there's a list index in the way somewhere)
    # these types don't get the full trace source as their source, but just the tracelet value (see get() below)
    type_classes.update({(category, type_name): type('.'.join(map(to_attr_name, (category, type_name))),
                                                     (TraceletTypeView, view_with_attrs(type_name, properties.keys())),
                                                     {})
                         for (category, type_name), properties in model_types.items()
                         if (category, type_name) in plural_types})

    # NB: the order of super classes is significant here, Trace needs the context argument, intrinsic_view (an
    #     AttrView) won't like it, method resolution order will make Trace.__init__ call intrinsic_view.__init__
    #     instead of DictView.__init__
    class TraceView(Trace, intrinsic_view):
        def __init__(self, source, context=None):
            super().__init__(source, context)

            # collect categorized types defined for this trace, regardless of origin
            types = set()
            for origin in origins:
                origin = self.get(origin) or {}
                for category_name, category in origin.items():
                    types.update({(category_name, type_name)
                                  for type_name in category.keys()
                                  # drop any non-defined (category, type) combo's (like (annotated, tags))
                                  if (category_name, type_name) in model_types.keys()})

            # set trace type attributes to self
            for category_name, type_name in types:
                type_class = type_classes[(category_name, type_name)]

                if (category_name, type_name) in mapped_types:
                    # gather the keys (typically data types) defined for this trace type in any origin
                    keys = set(chain.from_iterable((self.get('.'.join((origin, category_name, type_name))) or {}).keys()
                                                   for origin in origins))
                    type_attr = view_with_attrs(to_attr_name(type_name), keys)
                    setattr(self,
                            to_attr_name(type_name),
                            type_attr({
                                # include key name to type name to make TraceTypeView construct the right selection key
                                key: type_class(source,
                                                category_name, '.'.join((type_name, key)), origins,
                                                converters=model_types.get((category_name, type_name)))
                                for key in keys
                            }))
                elif (category_name, type_name) in plural_types:
                    # plural types don't automatically get an attribute
                    pass
                else:
                    setattr(self,
                            to_attr_name(type_name),
                            type_class(source,
                                       category_name, type_name, origins,
                                       converters=model_types.get((category_name, type_name))))

            # assign a set of type names to self
            self.types = {type_name for _, type_name in types}

        def get(self, key, default=None):
            # override DictView.get with a model-aware variant, allowing
            # 1. get('origin.category.type.property')
            # 2. get('category.type.property')
            # 3. get('type.property')
            # while not breaking
            # get('origin.category.type'), get('category.type'), get('category'), get('type'), get('origin.category'),
            # get('origin'), get('origin.type')
            steps = key.split(self._separator)

            candidate_origins = origins
            if steps[0] in origins and len(steps) >= 4:
                # explicit option 1, strip the origin from steps to hit conditions below while restricting the origins
                # that need to be checked
                candidate_origins = [steps.pop(0)]

            if steps[0] in types:
                # first key is a known type, insert its category (types is {name → category}) and move on
                steps.insert(0, types.get(steps[0]))

            if steps[0] in categories:
                # (new) first step is a category, return first available <origin>.<category>.<type>.<rest>
                # (depending on code above, there may be only one origin that needs to be checked)
                for origin in candidate_origins:
                    value = super().get(self._separator.join([origin] + steps), default=self._no_default)
                    if value is not self._no_default:
                        # value is available, figure out if it needs a type conversion
                        modeled = model_types.get((steps[0], steps[1]))
                        if modeled and len(steps) >= 3:
                            # determine the step that contains the actual property name
                            # (1 index further along for mapped types)
                            property_step = steps[2 if (steps[0], steps[1]) not in mapped_types else 3]
                            converter = modeled.get(property_step)
                            if converter:
                                # converter defined for value, apply
                                return converter(value)
                        if modeled and len(steps) == 2 and (steps[0], steps[1]) in plural_types:
                            # explicitly getting a plural type (on-trace / few), wrap with a view that does not take
                            # origins into account, accessing the local dict directly
                            type_class = type_classes[(steps[0], steps[1])]
                            return [type_class(element, steps[0], steps[1], modeled) for element in value]

                        return value

            # either option 1 or non-understandable or non-model key, use default implementation
            return super().get(key, default)

    type_doc = """Trace type attributes added from model: {}.

Trace type attributes are defined when a trace has that particular
trace type. Check for this with `'type_name' in trace.types`.""".format(
        ', '.join(to_attr_name(name) for _, name in model_types.keys())
    )
    TraceView.__doc__ = '{}\n\n{}'.format(Trace.__doc__, type_doc)

    return TraceView



# encoding=utf-8

from concurrent.futures import as_completed, FIRST_COMPLETED, ThreadPoolExecutor, wait
import csv
from functools import partial
from io import open  # this is essentially meaningless, but makes it possible to mock calls to open()
from os import makedirs, path
import re

from logbook import Logger

from hansken import fetch
from hansken.trace import expand_types


log = Logger(__name__)


def get_fields(trace, fields, prefix='system.extracted.', use_fallback=True, default=None):
    """
    Retrieves values for fields from source by calling
    source.get(prefix + field) for each field.

    :param trace: collection of mapped values
    :param fields: fields to retrieve a value for
    :param prefix: prefix used with get
    :param use_fallback: whether to try getting a field without prefix when
        a value for the full field name is not available
    :param default: value to use when no value was mapped
    :return: a dictionary with all of the requested fields and their value in
        the source of *trace*, or `None` if *trace* has no value for the field
    :rtype: `dict`
    """
    values = {}
    for field in fields:
        if isinstance(field, data_stream):
            log.info('retrieving {} stream for trace {}', field, trace.uid)
            values[field.field_name] = field.to_text(trace, default=default)
        else:
            value = trace.get('{prefix}{field}'.format(prefix=prefix or '', field=field), default=default)
            if value is None and use_fallback:
                # fallback to property without prefix only if allowed and value is None (0 or False are valid)
                value = trace.get(field, default=default)
            # collect
            values[field] = value

    return values


class type_fields:  # noqa: N801
    """
    An object to request that all fields of the requested types should be used
    as field names when exporting to CSV format. Multiple type names can be
    provided:

    .. code-block:: python

        to_csv(..., fields=type_fields('file', 'link'))
        # supplying two types at once is equivalent to supplying two separate type_fields
        to_csv(..., fields=[type_fields('file'), type_fields('link'))

    Both forms would result in headers like ``file.createdOn`` and
    ``link.target`` in the resulting export file.
    """
    # TODO: include_intrinsics, handle_list, handle_dict/map
    def __init__(self, *type_names):
        self.type_names = type_names


class data_stream:  # noqa: N801
    """
    An object to request text data for traces to be exported. Like
    `.type_fields`, this can be mixed with regular metadata fields:

    .. code-block:: python

        to_csv(..., fields=['data.raw.size', data_stream('raw', max_size=1024)])
        to_csv(..., fields=data_stream('text', max_size=None))

    A fallback text encoding can be provided for data stream exports that have
    no explicit or known text encoding:

    .. code-block:: python

        to_csv(..., fields=data_stream('plain', fallback_encoding='ascii')

    Encoding errors will result in replacement characters — ?'s — in the output
    CSV. The fallback encoding is unset by default, resulting in *no data* in
    the output CSV.

    For custom functionality (like including binary data encoded as hex or
    base64), this class can be extended. `.to_text` is called by the exporter
    (`.get_fields` when `.to_csv` is used) to get a `str` from a `.Trace` being
    exported.

    .. note::

        The maximum number of bytes that are read from the stream defaults to
        4KiB. While it's possible to include an entire data stream in the
        export, please note that data streams can grow quite large; use this
        with caution.

    .. note::

        As exporting data requires an additional HTTP request for each data
        stream of each `.Trace` being exported, including data in an export
        slows down the export considerably.
    """
    # data stream names that are known to contain utf-8 encoded text
    utf8_data_types = ('html', 'htmlText', 'ocr', 'text')
    # data streams other than those above might contain an explicit charset in their mime type
    encoding_pattern = re.compile(r"""charset=['"]?(?P<encoding>[\w\-]+)['"]?""")

    def __init__(self, data_type, max_size=4 << 10, fallback_encoding=None):
        self.data_type = data_type
        self.max_size = max_size
        self.fallback_encoding = fallback_encoding

    @property
    def field_name(self):
        return 'data.{}'.format(self.data_type)

    def get_encoding(self, trace):
        mime_type = trace.get('data.{}.mimeType'.format(self.data_type))
        encoding = self.encoding_pattern.search(mime_type) if mime_type else None
        if encoding:
            return encoding.group('encoding').lower()
        elif self.data_type in self.utf8_data_types:
            return 'utf-8'
        elif self.fallback_encoding:
            log.debug('falling back to encoding {} for stream {} of trace {}',
                      self.fallback_encoding, self.data_type, trace.uid)
            return self.fallback_encoding
        else:
            log.debug('unable to determine encoding of stream {} for trace {} (mime type: {})',
                      self.data_type, trace.uid, mime_type)
            return None

    def to_text(self, trace, default=None):
        """
        Get a `str` from a `.Trace`. Uses this `.data_stream`'s ``data_type``
        and ``max_size`` to retrieve the requested data and turns it into a
        `str` if possible.

        :param trace: the `.Trace` being exported
        :param default: the value to be used when retrieving the data or
            turning it into a `str` fails
        :return: a `str` representation of (a part of) the requested data
            stream
        """
        if self.data_type not in trace.data_types:
            return default

        encoding = self.get_encoding(trace)
        if not encoding:
            log.warn('omitting data for stream {} for trace {}, '
                     'unable to determine encoding and no fallback encoding provided', self.data_type, trace.uid)
            return default

        with trace.open(self.data_type, size=self.max_size) as data:
            data = data.read()
            return data.decode(encoding, errors='replace')


def to_csv(traces, output, fields,
           to_dict=get_fields,
           delimiter='\t',
           lineterminator='\n',
           encoding='utf-8',
           **fmtparams):
    """
    Writes values for fields from each trace in *traces* to *output*. Field
    names can be supplied as property names (e.g. ``'file.createdOn'``) or as
    `.type_fields` instances that automatically expand to properties defined
    for the specified types. Data can be included by using `.data_stream`
    instances.

    .. note::

        Using `.type_fields` instances **requires** that the *traces* argument
        carries the applicable trace model as attribute ``model``. Using
        `.data_stream` instances **requires** that the `.Trace` objects in the
        *traces* argument carry a `.ProjectContext` object. This is the case
        for `.SearchResult` instances obtained from calls like
        `.ProjectContext.search`:

        .. code-block:: python

            # obtain search results as normally
            results = context.search('query')
            # export the results to a local CSV file
            to_csv(results, 'path/to/export.csv',
                   # explicit fields and automatically expanded fields can be mixed
                   fields=['uid', 'name',
                           # include all model-defined metadata fields for email traces
                           type_fields('email'),
                           # include the first kilobyte of data from the plain data stream (converted to text)
                           data_stream('plain', max_size=1024)

        The exported file with have the explicitly provided fields like ``uid``
        and ``name``, but also fields from property names generated from the
        trace model retrieved from the `.ProjectContext`, like
        ``email.subject`` and ``email.to``. The `.data_stream` instance will
        cause a ``data.plain`` field.

    :param traces: collection of traces
    :param output: name of the file to write to
    :param fields: fields to retrieve values for, a sequence of property names
        (`str`) or `.type_fields` instances, used to generate field and
        property names from a trace type
    :param to_dict: callable to create a dict ``{field: value}``;
        passed kwargs: *trace*, *fields*
    :param delimiter: field delimiter in the output
    :param lineterminator: line terminator in the output
    :param encoding: text encoding for the output
    :param fmtparams: additional format parameters, see module csv in the
                      standard lib
    """
    def expand(fields):
        """
        Recursively resolves any instances of `type_fields` to the corresponding
        properties / field names and adds a field name to any `data_stream`
        instances. Yields *(name, field)* tuples, where the field name
        will only differ from the field for `data_stream` instances.
        """
        if isinstance(fields, str):
            # fields is just a field
            yield fields, fields
        elif isinstance(fields, data_stream):
            # fields is a data stream, handled in get_fields
            yield fields.field_name, fields
        elif isinstance(fields, type_fields):
            # user request to include all properties for a type
            for field in expand_types(traces.model, *fields.type_names):
                yield field, field
        else:
            # sequence of fields, resolve any type_fields instances
            for item in fields:
                yield from expand(item)

    # will use builtin csv module, open the file in text mode, *don't* translate line endings to os default
    # (by supplying lineterminator to the writer, a script would produce identical output on any os)
    output = open(output, 'w', newline='', encoding=encoding)

    # resolve any 'automatic headers' passed as type_fields instances, split fields names and field 'pointers'
    names, fields = zip(*expand(fields))
    with output:
        writer = csv.DictWriter(output,
                                fieldnames=names,
                                delimiter=delimiter,
                                lineterminator=lineterminator,
                                **fmtparams)
        writer.writeheader()
        for trace in traces:
            writer.writerow(to_dict(trace=trace, fields=fields))


def to_file(trace, output, stream='raw', offset=0, size=None, key=fetch, bufsize=1 << 20):
    """
    Writes a data stream of a trace to a file.

    :param trace: trace to write
    :param output: name of the file to write to
    :param stream: named stream to read from the trace
    :param offset: byte offset to start the stream on
    :param size: the number of bytes to make available
    :param key: key for the image of *trace* (default is to fetch the key
        automatically, if it's available)
    :param bufsize: buffer size to be used during the read/write loop
    """
    with open(output, 'wb') as out_file, trace.open(stream=stream, offset=offset, size=size, key=key) as stream:
        # pre-allocate a reusable buffer
        buffer = memoryview(bytearray(bufsize))
        # keep reading into buffer until we reach the end of the stream
        num_read = stream.readinto(buffer)
        while num_read:
            out_file.write(buffer[:num_read])  # only write up to num_read, we don't want trailing junk
            num_read = stream.readinto(buffer)


def safe_name(trace, num=None, split=1000, stream='raw', template='{trace.image_id}_{trace.id}_{stream}_{trace.name}'):
    """
    Generate a file name for a trace. Resulting file name can contain unicode
    characters, but no slashes, backslashes or line endings.

    :param trace: trace to name
    :param num: the number of this trace within the set being exported
    :param split: the max number of files in a directory
    :param stream: the named stream to be exported
    :param template: format string used as the file name, slashes and newlines
        are replaced by underscores in the result;
        passed kwargs: *trace*, *num*, *split*, *stream*
    :return: generated file name
    """
    file_name = template.format(trace=trace,
                                num=num,
                                split=split,
                                stream=stream)
    # NB: unicode characters in trace names are assumed to be safe
    return re.sub(r'[/\\\n\r]', '_', file_name.strip())


def bulk(traces, dest, split=1000, stream='raw', fname=safe_name, write=to_file, on_error=None, side_effect=None,
         jobs=16):
    """
    Performs a bulk export of traces to dest.

    .. note::

        `.bulk` is internally parallellized by default, **requiring** that the
        argument to *write* is thread-safe. *safe_name*, *on_error* and
        *side_effect* are all called from the calling thread after the export
        of a particular trace in *traces* was processed.

        As *on_error* is **not** called from the ``except`` clause that catches
        the `Exception` instance, logging the exception with its traceback
        requires special care to pass the ``exc_info`` keyword to either
        `logging` or `logbook`. Leaving *on_error* as `None` will raise a
        `ValueError` on the thread calling `bulk` with the error that is
        processed first as its cause.

        This also means that the order of traces with which *write*, *on_error*
        and *side_effect* are called need not be the same order as that of
        *traces*. To turn this parallellism off, pass ``jobs=False``.

    :param traces: collection of traces to export
    :param dest: path to export traces to
    :param split: max number of files per directory (when set to `None`, all
        files will be saved to the same directory)
    :param stream: stream name to read from the traces, optionally supplied as
        a `callable` returning the stream name (*trace* will be omitted from
        export if the return value is falsy);
        passed kwargs: *trace*
    :param fname: `callable` to generate a file name for a trace;
        passed kwargs: *trace*, *num*, *split*, *stream* (defaults to
        `.safe_name`, resulting in a file name that uses both
        ``trace.image_id`` and ``trace.id``, ensuring the name is unique
        within a project)
    :param write: *thread-safe* `callable` to write a trace to a file name;
        passed kwargs: *trace*, *output*, *stream* (defaults to `.to_file`)
    :param on_error: `callable` to report an error thrown during ``write()``;
        passed kwargs: *num*, *trace*, *stream*, *output*, *exception*
    :param side_effect: `callable` to perform a side effect for each exported
        trace;
        passed kwargs: *trace*, *stream*, *num*, *split*, *dest*, *folder*,
        *file*, *output*
    :param jobs: maximum number of data exports to run in parallel (an `int`),
        or `False` to turn parallel processing of *traces* off
    :raises ValueError: on the first error result when *on_error* is not
        supplied (the error is set as the cause)
    """
    if jobs is True or not isinstance(jobs, int):
        raise TypeError('an int or False is required for arguments jobs')

    num_total = len(traces)
    # determine the ideal length of the folder names for split sets
    order = len(str(num_total // split)) if split else 1

    try:
        executor = ThreadPoolExecutor(max_workers=jobs or 1, thread_name_prefix='bulk_export_worker')
    except TypeError:
        # thread_name_prefix is not supported for Python 3.5 (backport for 2.7 *does* have this argument)
        executor = ThreadPoolExecutor(max_workers=jobs or 1)

    with executor:
        # track pending tasks as a set of futures
        futures = set()
        # allow a fixed number of tasks to be pending
        pending_threshold = jobs * 2 if jobs else 1
        # prepare a call to run_task needing only a task
        target = partial(run_task, write)

        for num, trace in enumerate(traces):
            # resolve stream if caller wants to decide which one to select
            selected_stream = stream(trace=trace) if callable(stream) else stream
            if not selected_stream:
                # user-supplied stream selector returned a falsy value, that won't be something we can export
                # drop this trace from the export tasks, but warn about it
                log.warn('omitting trace {} from bulk export, no stream to export was selected', trace.get('uid'))
                continue

            # determine folder number only if a split is requested
            folder = num // split if split else None
            file = fname(trace=trace, num=num, split=split, stream=selected_stream)

            if folder is not None:
                folder = path.join(dest, '{folder:0{len}}'.format(folder=folder, len=order))
            else:
                # no need to insert a split number folder
                folder = dest

            makedirs(folder, exist_ok=True)

            output = path.join(folder, file)

            # define task as a collection of locals needed by write to complete the task
            task = {
                'trace': trace, 'stream': selected_stream,
                'num': num, 'split': split,
                'dest': dest, 'folder': folder, 'file': file, 'output': output,
            }
            # submit the execution of task to the executor, tracking the future for … future use
            futures.add(executor.submit(target, task))

            # 'periodically' check how many tasks are currently tracked
            # (this provides 'back pressure' on consuming traces and avoids flooding the executor queue)
            if len(futures) >= pending_threshold:
                # retrieve the futures that are already done or wait for the first one to be done
                done, _ = wait(futures, return_when=FIRST_COMPLETED)
                # untrack the done futures (effectively creating len(done) space for new tasks to be submitted)
                futures.difference_update(done)

                # process the results of all the done tasks
                for future in done:
                    task, error = future.result()
                    process_result(task, error, on_error, side_effect)

        # traces has been consumed, all tasks submitted, wait for and process all tasks still pending
        for future in as_completed(futures):
            task, error = future.result()
            process_result(task, error, on_error, side_effect)


def run_task(write, task):
    """
    Runs *task* using *write*.

    *(Intended for internal use with `.bulk`)*

    :param write: *thread-safe* callable to write a trace to a file name;
        passed kwargs: *trace*, *output*, *stream* (defaults to `.to_file`)
    :param task: the task to perform using *write*
    :return: a 2-tuple *(original task, optional error)*
    """
    try:
        # save trace' stream to the generated output path
        write(trace=task['trace'], output=task['output'], stream=task['stream'])
        return task, None
    except Exception as e:
        return task, e


def process_result(task, error, on_error, side_effect):
    """
    Processes the result of *task*, calling *side_effect* on success,
    *on_error* on failure.

    *(Intended for internal use with `.bulk`)*

    :param task: the task that was performed
    :param error: an optional error encountered while performing *task*
    :param on_error: callable to report an error thrown during ``write()``;
        passed kwargs: *num*, *trace*, *stream*, *output*, *exception*
    :param side_effect: callable to perform a side effect for each exported
        trace;
        passed kwargs: *trace*, *stream*, *num*, *split*, *dest*, *folder*,
        *file*, *output*
    :raise ValueError: when *error* is provided and *on_error* is not (note
        that an error raised by *side_effect* counts as *error* being provided)
    """
    if not error and callable(side_effect):
        try:
            # call the side effect with a lot of info:
            side_effect(
                # info on the stream being exported
                trace=task['trace'], stream=task['stream'],
                # split info
                num=task['num'], split=task['split'],
                # destination path (and parts)
                dest=task['dest'], folder=task['folder'], file=task['file'], output=task['output']
            )
        except Exception as e:
            error = e

    if error:
        if callable(on_error):
            on_error(num=task['num'], trace=task['trace'], stream=task['stream'], output=task['output'],
                     exception=error)
        else:
            raise ValueError('export or side effect of trace {} failed'.format(task['trace'].get('uid'))) from error



# encoding=utf-8

from argparse import ArgumentParser, Namespace, SUPPRESS
from datetime import datetime, timezone, tzinfo
from getpass import getuser
import os
import shlex
import warnings

from dateutil.tz import gettz
from logbook import DEBUG, INFO, WARNING
from logbook import FileHandler, Logger, NullHandler, set_datetime_format, StderrHandler
from logbook.compat import redirected_logging, redirected_warnings
import requests

from hansken import __version__, envvars
from hansken.auth import HanskenAuthBase, resolve
from hansken.remote import Connection, MultiProjectContext, ProjectContext
from hansken.util import MultiContext

try:
    # ensure urllib3 is a module-level name for testability
    import urllib3
except ImportError:
    urllib3 = None


log = Logger(__name__)


def from_envvar(envvar, provided=None):
    if provided:
        return provided

    return os.environ.get(envvar, None)


def prompt_user():
    try:
        user = getuser()
        read = input('username [{}]: '.format(user))
        if read.strip():
            user = read
    except (KeyboardInterrupt, EOFError):
        user = None
    except Exception:  # documentation of getuser: "otherwise, an exception is raised"
        user = input('username: ')

    return user


# define a parent parser to define global options
parent = ArgumentParser(add_help=False)
parent.add_argument('-v', '--verbose', action='count', help='be verbose')
parent.add_argument('-l', '--log', metavar='FILE',
                    default=from_envvar(envvars.log),
                    help='log messages to FILE (use - for standard error, '
                         'log messages are hidden by default)')
parent.add_argument('-z', '--timezone', metavar='TZ', type=gettz, default=timezone.utc,
                    help='log messages with timestamps in timezone TZ (e.g. Europe/Amsterdam), '
                         'defaults to UTC')
# default endpoint and project to the value in env
parent.add_argument('-e', '--endpoint', default=from_envvar(envvars.endpoint),
                    help='connect to the Hansken REST endpoint at ENDPOINT '
                         '(defaults to environment variable {})'.format(envvars.endpoint))
parent.add_argument('--keystore', metavar='ENDPOINT',
                    default=from_envvar(envvars.keystore),
                    help='connect to the Hansken keystore REST endpoint at ENDPOINT '
                         '(defaults to environment variable {})'.format(envvars.keystore))
parent.add_argument('--preference', metavar='ENDPOINT',
                    default=from_envvar(envvars.preference),
                    help='connect to the Hansken preference REST endpoint at ENDPOINT '
                         '(defaults to environment variable {})'.format(envvars.preference))
parent.add_argument('--idp', metavar='IDP_ID',
                    default=from_envvar(envvars.idp),
                    help='select SAML identity provider named IDP_ID if available '
                         '(defaults to environment variable {})'.format(envvars.idp))
parent.add_argument('--idp-url', metavar='ENDPOINT',
                    default=from_envvar(envvars.idp_url),
                    help=SUPPRESS)
parent.add_argument('--idp-realm', metavar='REALM',
                    default=from_envvar(envvars.idp_realm),
                    help=SUPPRESS)
parent.add_argument('--sso-url', metavar='ENDPOINT',
                    default=from_envvar(envvars.sso_url),
                    help=SUPPRESS)
parent.add_argument('-u', '--user', '--username', dest='username', default=None,
                    help='specify a username (will otherwise be prompted for if needed)')
parent.add_argument('-U', '--prompt-username', action='store_const', const=prompt_user, dest='username',
                    help=SUPPRESS)
parent.add_argument('--password', dest='password', default=None,
                    help=SUPPRESS)
parent.add_argument('--connection-pool-size', type=int, default=None,
                    help='override default HTTP(S) connection pool size')
# add an argument to disable ssl/tls certificate verification
parent.add_argument('--no-verify', dest='verify', action='store_false', default=True,
                    help='disable certification verification for secure connections, use with caution')

# NB: add a hidden hdfs configuration option for operators, allowing a single argument file per Hansken instance
# (value for hdfs_url is ignored for most commands, can be overridden by -o/--output for upload command)
parent.add_argument('--hdfs-url', dest='hdfs_url', help=SUPPRESS)

# password and auth are not really arguments, but used in combo with user/idp/sso; make sure they're defined
parent.set_defaults(password=None, auth=None)


# define the actual parser to be used
# (defined as a local class to override result for and repr() in doc as default for arg)
class hansken_parser(ArgumentParser):  # noqa: N801
    def __repr__(self):
        return '<hansken.py default parser>'


hansken_parser = hansken_parser(fromfile_prefix_chars='@')
# define the version switch only on the global parser
hansken_parser.add_argument('--version', action='version', version=__version__)
# parse arguments like a shell when reading args from file
# (the default effectively only splits lines)
hansken_parser.convert_arg_line_to_args = lambda line: shlex.split(line, comments=True)

# enable subcommands, saving the name of the command in argument command
sub_commands = hansken_parser.add_subparsers(dest='command', metavar='COMMAND')
sub_commands.required = True  # 'fix' for required=True not being allowed in add_subparsers above


def create_parser(requires_project=False, **kwargs):
    warnings.warn('hansken.tool.create_parser has been renamed to hansken.tool.create_argument_parser',
                  DeprecationWarning)
    return create_argument_parser(requires_project=requires_project, **kwargs)


def create_argument_parser(requires_project=False, **kwargs):
    """
    Creates an `ArgumentParser` to be used with `.run` alongside
    ``with_context`` or ``with_admin``. This parser's main feature is to *not*
    add a ``-h/--help`` argument, which is attached to the main argument
    parser used by `.run`.

    :param requires_project: whether the parser should be fitted with a
        required positional argument ``project``
    :param kwargs: any `ArgumentParser` constructor arguments (see `argparse`
        documentation)
    :return: an `ArgumentParser` that won't clash with ``hansken.py``'s parent
        parser
    :rtype: `argparse.ArgumentParser`
    """
    # don't add help, returned parser will serve as a parent that will itself add help
    if kwargs.pop('add_help', False):
        raise ValueError('parser cannot specify -h/--help')

    parser = ArgumentParser(add_help=False, **kwargs)
    if requires_project:
        parser.add_argument('project', metavar='PROJECT',
                            help='attach to project with id PROJECT')
    return parser


def add_command(name, command, optional_project=True, **kwargs):
    """
    Adds a named command to be run when specified on the command line. Use the
    return value to add additional arguments for the newly added command.

    :param name: name of the command to add
    :param command: callable that runs the actual command, first positional
        argument to command is the `argparse.Namespace` object created by
        `argparse`.
    :param optional_project: whether to include a -p/--project option
    :param kwargs: any `ArgumentParser` constructor arguments (see `argparse`
        documentation)
    :return: an `ArgumentParser` for the newly added command
    :rtype: `argparse.ArgumentParser`
    """
    # add a new parser to sub commands, pass along the parent (enables command -e http://...) and any kwargs
    sub_parser = sub_commands.add_parser(name, parents=[parent], **kwargs)
    # set the callable as the actual command to be run
    sub_parser.set_defaults(run_command=command)

    if optional_project:
        # only add the optional -p/--project if sub command needs it
        sub_parser.add_argument('-p', '--project', default=from_envvar(envvars.project),
                                help='attach to project with id PROJECT '
                                     '(defaults to environment variable {})'.format(envvars.project))
    return sub_parser


def set_command(command, parents=None, **kwargs):
    """
    Overwrite the default functionality of the ``hansken.py`` command line
    interface. Note that this method should be called only once, either from
    user code, passing the resulting `ArgumentParser` to `.run` or by `.run`
    itself when used with one of the *with_\\** keyword arguments.

    :param command: a `callable` to be called with a single positional
        argument of type `argparse.Namespace` after parsing arguments
    :param parents: a sequence of parent parsers to use aside from the always
        inserted parent parser defining the global configuration options
    :param kwargs: any `ArgumentParser` constructor arguments (see `argparse`
        documentation)
    :return: an `ArgumentParser` for the new command
    :rtype: `argparse.ArgumentParser`
    """
    if hasattr(set_command, 'called') and set_command.called:
        raise ValueError('cannot set main command more than once')

    if not callable(command):
        raise ValueError('command must be callable')

    # 'copy' behaviour of global parser to the new one
    fromfile_chars = kwargs.pop('fromfile_prefix_chars', '@')
    fromfile_splitter = kwargs.pop('convert_arg_line_to_args', hansken_parser.convert_arg_line_to_args)

    # add the global parent to any parents provided by the caller
    parents = list(parents or [])
    parents.append(parent)
    # define a new default nameless command (the empty name will make --help appear as if there's no sub command)
    sub_parser = sub_commands.add_parser('', parents=parents, fromfile_prefix_chars=fromfile_chars, **kwargs)
    sub_parser.convert_arg_line_to_args = fromfile_splitter
    sub_parser.set_defaults(command='', run_command=command)

    # mark this function as called, subsequent calls produce confusing behaviour and should fail
    setattr(set_command, 'called', True)
    # provide the sub parser to the caller, let them add additional arguments
    return sub_parser


def resolve_logging(args):
    """
    Creates `logbook.Handler` instances from available parameters.

    :param args: an `argparse.Namespace` object with parsed commandline
        arguments
    :return: a context manager within which logging is configured from the
        provided command line arguments
    """
    levels = [WARNING, INFO, DEBUG]
    verbosity = args.verbose
    if verbosity not in levels:
        # make sure verbosity is an int (default None)
        verbosity = max(0, min(verbosity or 0, len(levels) - 1))
        verbosity = levels[verbosity]

    if args.log:
        tz = args.timezone
        if not isinstance(tz, tzinfo):
            # should timezone be provided through run, argument won't have been passed to gettz
            tz = gettz(tz)

        def utc_dt():
            return datetime.now(tz=tz)

        # force logbook to use tz-aware timestamps, log in requested timezone (default UTC)
        set_datetime_format(utc_dt)

        # when logging is actively used, also steal both logging.* and warnings.warn calls
        # bind the logbook handlers application-wide
        if args.log == '-':
            return MultiContext(redirected_logging(),
                                redirected_warnings(),
                                StderrHandler(level=verbosity).applicationbound())
        elif args.log:
            return MultiContext(redirected_logging(),
                                redirected_warnings(),
                                FileHandler(args.log, level=verbosity).applicationbound())
    else:
        # no dest, no real handler
        return NullHandler().applicationbound()


def resolve_auth(args):
    """
    Creates a HanskenAuthBase instance from available parameters.

    - endpoint + idp        -> select named IDP from REST API, probe for
                               implementation to use
    - idp_url + sso_url     -> OpenAM auth with Kerberos SSO
    - idp_url + user + pass -> OpenAM auth using REST login
    - user + pass           -> SAML proxy auth using REST login
    - otherwise             -> None

    :param args: an argparse.Namespace object with parsed commandline arguments
    :return: a HanskenAuthBase instance if args contained parameters that
             enable one of the options, None otherwise
    """
    return resolve(args.username, args.password,
                   base_url=args.endpoint, idp_id=args.idp,
                   idp_url=args.idp_url, idp_realm=args.idp_realm, sso_url=args.sso_url,
                   verify=args.verify)


def _disable_insecure_request_warning():
    try:
        # attempt to disable repeated verbose warnings when certificate verification is explicitly turned off
        # (remove when hansken.py depends on requests>=2.16.0)
        requests.packages.urllib3.disable_warnings(requests.packages.urllib3.exceptions.InsecureRequestWarning)
        return True
    except Exception:  # nosec
        pass

    try:
        # newer installations will fail the above on the 'packages' module being absent, fall back to the 'new way'
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        return True
    except Exception:  # nosec
        pass

    # disabling the warning failed
    return False


def run(args=None, using_parser=hansken_parser, with_connection=None, with_context=None,
        with_multi_context=None, error='raise', with_admin=None, **defaults):
    """
    Run the command line tool, either as the command line interface defined by
    ``hansken.py`` itself, or running user code, piggybacking on all the
    command line argument parsing and resolving done by ``hansken.py``.

    .. note::

        `.run` parses the command line by default. If user code does its own
        command line parsing, either pass applicable *args* to `.run` or use
        `.set_command`, passing the resulting `argparse.ArgumentParser` to
        `.run` as *using_parser* after adding additional user arguments.

    .. note::

        By default, `.run` calls `exit` on the `argparse.ArgumentParser` in
        case of an error raised from whatever command is called (this includes
        user-provided commands). This call results in a CLI-like error message
        containing the exception message of the error that was caught. As the
        call to `exit` raises a `SystemExit`, any user code *after* the call
        to `.run` will not be executed. When the *error* argument to `.run` is
        set to ``'raise'`` or the arguments parser sets verbose mode, `exit`
        is *not* called, the exception is reraised instead.

    :param args: arguments to the tool, a `list` of `str`, or `None`, in which
        case *args* will be read from the command line
    :param using_parser: the `argparse.ArgumentParser` to be used, either the
        parser that processes the command line interface defined by
        ``hansken.py`` itself or one that is the result of `.set_command`
    :param with_connection: a user-defined `callable` to be run with a
        `.Connection` instance, passed as keyword argument ``connection``
    :param with_context: a user-defined `callable` to be run with a
        `.ProjectContext` instance, passed as keyword argument ``context``
    :param with_multi_context: a user-defined `callable` to be run with a
        `.MultiProjectContext` instance, passed as keyword argument ``context``
    :param error: how to deal with an error raised from the selected command
        runner, defaults to propagating errors to the caller of `.run`, set to
        ``'exit'`` to call ``using_parser.exit()`` with a non-zero return code
        and failure message
    :param defaults: named arguments to be used as defaults when parsing the
        command line, matching the destinations of defined arguments (e.g.
        ``project='a-project-uuid'``)
    :return: the return value of the command that was ultimately called (when
        `.run` was called with either *with_context* or *with_admin*, the
        return value of that callable will be returned by `.run`)
    """
    def has_project(parser):
        # try to determine whether parser has already defined either default or argument for a project
        if parser.get_default('project'):
            return True

        return any(action.dest == 'project' for action in parser._actions)

    # run with either connection or context, not both
    if (with_context, with_connection, with_admin).count(None) < 2:
        raise ValueError('use either with_context=<my_func> or with_connection=<my_func>, not both')

    # check whether the parser being used is a custom one
    custom_parser = using_parser is not hansken_parser

    # user requests to be called with a context, overwrite our main command (…)
    if with_context:
        def run(args):
            # (…) with something that creates a ProjectContext from the env/command line (…)
            context = ProjectContext(args.endpoint, args.project, args.keystore, args.preference,
                                     auth=args.auth, connection_pool_size=args.connection_pool_size,
                                     verify=args.verify)

            kwargs = {'context': context}
            if custom_parser:
                # (optionally add the parsed args to the passed arguments)
                kwargs.update(args=args)

            # (…) and calls the user's function with that context
            return with_context(**kwargs)

        # set the parser to be used to the one resulting in the user's function being called
        # add the custom parser as one of the parents of the resulting one to incorporate parsed user args
        using_parser = set_command(run, parents=[using_parser] if custom_parser else None)
        if not defaults.get('project') and not has_project(using_parser):
            # user did not supply an explicit project id to run(), we'll require it form the command line
            using_parser.add_argument('project', metavar='PROJECT',
                                      help='attach to project with id PROJECT')

    if with_multi_context:
        def run(args):
            # (…) with something that creates a MultiProjectContext from the env/command line (…)
            context = MultiProjectContext(args.endpoint, args.project, args.keystore, args.preference,
                                          auth=args.auth, connection_pool_size=args.connection_pool_size,
                                          verify=args.verify)

            kwargs = {'context': context}
            if custom_parser:
                # (optionally add the parsed args to the passed arguments)
                kwargs.update(args=args)

            # (…) and calls the user's function with that context
            return with_multi_context(**kwargs)

        # set the parser to be used to the one resulting in the user's function being called
        # add the custom parser as one of the parents of the resulting one to incorporate parsed user args
        using_parser = set_command(run, parents=[using_parser] if custom_parser else None)
        if not defaults.get('project') and not has_project(using_parser):
            # user did not supply an explicit project id to run(), we'll require it form the command line
            using_parser.add_argument('project', metavar='PROJECT', nargs='+',
                                      help='attach to project with id PROJECT')

    connection_kwarg = 'connection'
    if with_admin:
        warnings.warn('hansken.tool.run(with_admin=…) has been deprecated, use hansken.tool.run(with_connection=…) '
                      'instead', DeprecationWarning)
        connection_kwarg = 'admin'

    connection_callable = with_connection or with_admin
    # user requests to be called with connection, overwrite our main command (…)
    if connection_callable:
        def run(args):
            # (…) with something that creates a Connection from the env/command line (…)
            connection = Connection(args.endpoint, args.keystore, args.preference,
                                    auth=args.auth, connection_pool_size=args.connection_pool_size,
                                    verify=args.verify)

            kwargs = {connection_kwarg: connection}
            if custom_parser:
                # (optionally add the parsed args to the passed arguments)
                kwargs.update(args=args)

            # (…) and calls the user's function with that connection
            return connection_callable(**kwargs)
        # set the parser to be used to the one resulting in the user's function being called
        # add the custom parser as one of the parents of the resulting one to incorporate parsed user args
        using_parser = set_command(run, parents=[using_parser] if custom_parser else None)

    # parse args (defaults to sys.argv[1:]) using hansken_parser or something resulting in a user function being
    # called, exiting here if not satisfiable
    # pre-create the Namespace where arguments are stored, initialized to user defaults
    args = using_parser.parse_args(args, Namespace(**defaults))

    # create a log handler from parsed arguments, use it while running command
    with resolve_logging(args):
        # when certificate validation is disabled, attempt to disable the insecure warning (…)
        if not args.verify and not _disable_insecure_request_warning():
            # (…) but log when that fails
            log.warn('failed to disable InsecureRequestWarning in 3rd party package')

        if not isinstance(args.auth, HanskenAuthBase):
            # allow auth to be 'turned off' if explicitly set to False
            args.auth = None if args.auth is False else resolve_auth(args)

        try:
            log.info('initial argument handling complete, hansken.py version {}', __version__)
            # run action with the parsed arguments
            return args.run_command(args)
        except Exception as e:
            if error == 'exit' and not args.verbose:
                # calls to run with_* will likely have an empty command string, fall back to the parser's prog
                command = args.command or using_parser.prog
                command = command.strip()

                using_parser.exit(1, '\n{} failed: {}\n'.format(command, str(e)))
            else:
                # error == 'raise' or verbose mode, propagate error
                raise


# force adding commands from command sub-modules
# (note that a from ... import ... can't be used here as that would cause a circular import)
import hansken.tool.command_backup  # noqa: E402,F401,I100,I202
import hansken.tool.command_extract  # noqa: E402,F401
import hansken.tool.command_grant  # noqa: E402,F401
import hansken.tool.command_mount  # noqa: E402,F401
import hansken.tool.command_quickstart  # noqa: E402,F401
import hansken.tool.command_shell  # noqa: E402,F401
import hansken.tool.command_stats  # noqa: E402,F401
import hansken.tool.command_tasks  # noqa: E402,F401
import hansken.tool.command_tools  # noqa: E402,F401
import hansken.tool.command_upload  # noqa: E402,F401
import hansken.tool.command_versions  # noqa: E402,F401



# encoding=utf-8

from itertools import chain
from os import path

from setuptools import find_packages, setup


# read version number from file
here = path.dirname(__file__)
with open(path.join(here, 'hansken', 'VERSION')) as version:
    version_string = version.read().strip()

with open(path.join(here, 'README.md')) as readme:
    long_description = readme.read()

dependencies = ['decorator',
                'ijson>=3.1',
                'iso8601',
                'logbook>=1.0',
                'more-itertools',
                'python-dateutil',
                'requests>=2.7.0',
                'tabulate']

extras = {
    'mount': ['fusepy'],
    'kerberos': ['requests-kerberos'],
    'report': ['jinja2', 'weasyprint'],
    'shell': ['ipython'],
}

# add an additional 'extra' to encompass every optional feature
extras['all'] = list(chain.from_iterable(extras.values()))

setup(
    name='hansken',
    version=version_string,
    author='Netherlands Forensic Institute',
    author_email='hansken-support@nfi.nl',
    url='https://hansken.org/',
    description='Python API to the Hansken REST endpoint',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],

    packages=find_packages(),
    include_package_data=True,
    install_requires=dependencies,
    extras_require=extras,
    entry_points={
        'console_scripts': ['hansken=hansken.tool:run']
    }
)
